import json
from datetime import date, datetime, timedelta
from decimal import Decimal
from enum import Enum
from typing import Any
from uuid import UUID

from . import enums
from .errors import ResponseFormatError

ENUMS = {
    'cls_email_logs': enums.ClassEmailLogs,
    'cls_lab_limit': enums.ClassLabLimit,
    'acc_type': enums.AccountType,
    'pod_cat': enums.PodCategory,
    'pod_admin_state': enums.PodAdminState,
    'pod_current_state': enums.PodCurrentState,
    'pc_icon': enums.PCIcon,
    'pc_type': enums.PCType,
    'pl_icon': enums.PLIcon,
    'vh_com_path': enums.VirtualHostComPath,
    'vm_role': enums.VirtualMachineRole,
    'res_type': enums.ReservationType,
    'date_format': enums.DateFormat,
    'time_format': enums.TimeFormat,
    'cls_ext_slots_per_res': (enums.ClassExtensionSlots, int),
    'com_ext_slots_per_res': (enums.CommunityExtensionSlots, int),
    'con_access': enums.ContentAccessiblity,
    'free_alarm': enums.FreeAlarm,
}

ENUM_CSV = {
    'acc_privs': enums.AccountPrivileges,

    # Pod
    'pod_cat_values': enums.PodCategory,
}

STR_CSV = {
    'con_pt_ids',
}

DATE_FIELDS = {
    # Class
    'cls_end_date', 'cls_start_date',

    # System
    'sys_lic_exp_date', 'sys_maint_ends', 'sys_sdn_release_date',

    # Reservation
    'local_ymd',

    # History
    'crs_release_date',
}

DATETIME_FIELDS = {
    # HDR
    'time', 'hdr_time',

    # User
    'acc_last_login', 'acc_session_time_started', 'acc_time_created', 'acc_time_last_access',

    # Pod
    'time', 'vh_bios_date', 'vh_date_modified', 'vh_date_tested', 'vm_date_added', 'vh_date_added',
    'vm_date_added',

    # Reservation
    'res_active_time', 'res_attend_time', 'res_end', 'res_post_time', 'res_preload_time',
    'res_start', 'utc', 'local', 'time_in_local', 'time_in_utc', 'time_out_local', 'time_out_utc',
    'res_end_utc', 'res_start_utc', 'start_time', 'end_time',

    # History
    'lh_active_time', 'lh_attend_time', 'lh_end_time', 'lh_start_time', 'res_active_time', 'res_attend_time',
    'res_end_actual', 'res_end_sched',
    'res_load_time', 'res_post_time', 'res_save_time', 'res_start', 'res_term_time', 'rh_report_time'

    # VM
    'vdc_date_added', 'vdc_date_tested', 'vh_bios_date', 'vh_date_added', 'vh_date_modified', 'vh_date_tested',
    'vm_date_added', 'vhp_time',

    # System
    'sp_time',
}

TIMEDELTA_FIELDS = {
    # Reservation
    'res_remaining_dhms',
}

BOOLEAN_FIELDS = {
    # Class
    'cls_change_ex', 'cls_no_delete', 'cls_retain_ilt', 'cls_retain_st', 'cls_self_sched', 'cls_team_sched', 'lead',
    'con_managed', 'con_global', 'registered', 'auth_all_communities',

    # System
    'sys_logins_enabled',

    # User
    'acc_can_login', 'acc_pw_change', 'acc_sys', 'test_taken', 'com_enabled',
    'enabled', 'logout', 'changed',
    'acc_nlx_terms', 'acc_npd_terms',

    # Pod
    'pod_acl_enabled', 'pod_auto_net_enabled', 'pod_auto_net_host_setup', 'pod_auto_net_host_teardown',
    'pod_managed', 'pt_removable', 'pc_online', 'pc_revert_to_scrub', 'pc_vnc_use_copy_rect',
    'vh_online', 'vh_pra_enabled', 'vm_auto_display', 'vm_auto_network', 'vm_auto_settings', 'vm_sanity_checks',
    'pc_revert_to_scrub', 'pc_vnc_use_copy_rect', 'vh_pra_enabled', 'vm_auto_display', 'vm_auto_network',
    'vm_auto_settings', 'vm_sanity_checks', 'pt_removable', 'pod_managed', 'pod_dyn_vlan'

    # Reservation
    'res_done', 'res_is_active', 'tz_is_dst',

    # History
    'res_done', 'rh_reported'

    # VM
    'notify_progress', 'notify_complete', 'vh_online', 'vh_pra_enabled', 'data_available'

    # Paging
    'out_of_bounds',
}

DECIMAL_FIELDS = {
    # System
    'gnice_pct', 'guest_pct', 'idle_pct', 'iowait_pct', 'irq_pct', 'nice_pct', 'soft_pct',
    'steal_pct', 'sys_pct', 'usr_pct', 'uptime_sec', 'used_pct', 'sp_value',
}

INT_FIELDS = {
    # Class
    'cls_max_slots_per_res', 'cls_min_hours_btw_res', 'cls_retain_period', 'com_id' 'cls_id', 'acc_id',
    'con_trustee',

    # Lab
    'con_build', 'con_fs_id', 'con_trustee', 'ex_index', 'ex_minutes', 'ex_fs_id',

    # Pod
    'acc_id', 'pod_id', 'pod_csw_base_port', 'pod_csw_base_vlan', 'pod_csw_id', 'pod_id', 'pod_id_avail_count',
    'pod_id_base', 'pod_id_range_min', 'pod_id_range_max', 'pod_osm_enabled', 'non_osm_pod_count', 'pod_vlan_pool_size',
    'sys_total_pod_count', 'clone_vh_id', 'pl_index', 'dev_id', 'line', 'pc_id', 'pev_id', 'pod_id', 'res_id',
    'severity', 'device_count', 'pt_apc_port_count', 'pt_as_port_count', 'pt_build', 'pt_csw_port_count', 'pt_pod_max',
    'pt_vlan_pool', 'remote_pc_count', 'pod_res_id', 'pc_pod_id', 'pc_pod_index', 'pc_vnc_port', 'pl_index', 'vdc_id',
    'vh_cpu_cores', 'vh_cpu_mhz', 'vh_cpu_threads', 'vh_cpu_threads', 'vh_id', 'vh_memory_mb', 'vh_pra_max_cpu',
    'vh_pra_max_mem_mb', 'vh_pra_max_vm', 'vhg_id', 'vm_alloc_cpu_n', 'vm_alloc_mem_mb', 'vm_child_count', 'vm_id',
    'vm_parent_id', 'vm_runtime_vh_id', 'pt_apc_port_count', 'pt_as_port_count', 'pt_build', 'pt_csw_port_count',
    'pt_index', 'pt_pod_max', 'pt_vlan_pool', 'remote_pc_count', 'pt_apc_port_count', 'pt_as_port_count',
    'pt_csw_port_count', 'pt_index', 'pt_pod_max', 'pt_vlan_pool', 'remote_pc_count',

    # Reservation
    'acc_com_id', 'acc_id', 'cls_div_id', 'cls_id', 'pod_id', 'res_acc_id', 'res_cls_id', 'res_com_id', 'res_flags',
    'res_id', 'res_init_fail', 'res_minutes', 'res_pod_id', 'res_remaining_sec', 'res_id', 'res_minutes', 'serial',
    'slot_minutes', 'acc_com_id', 'cls_div_id', 'cls_id', 'pod_id', 'res_acc_id', 'res_cls_id', 'res_com_id',
    'res_flags', 'res_init_fail', 'res_minutes', 'res_pod_id', 'active_reservations', 'completed_reservations',
    'future_reservations', 'pods_in_use', 'total_reservations', 'days', 'hours', 'minutes', 'months', 'local_day',
    'local_day_of_week', 'local_hour_12', 'local_hour_24', 'local_minute', 'local_second', 'local_year', 'tz_id',
    'tz_id_in', 'tz_id_out',

    # History
    'acc_id', 'cli_bytes_in', 'cli_bytes_out', 'cli_connects', 'cli_seconds', 'cls_id', 'com_id', 'lh_attend_min',
    'lh_n_users', 'vnc_bytes_in', 'vnc_bytes_out', 'vnc_connects', 'vnc_seconds', 'pod_id', 'res_extensions', 'res_id',
    'rh_attend_min', 'rh_init_fail_total', 'rh_n_lab_devices', 'rh_n_users', 'rh_n_vcpu', 'rh_n_vm', 'rh_n_vm_hosts',
    'rh_resume_total', 'rh_suspend_total'

    # System
    'cpu_n', 'free_b', 'avail_b', 'total_b', 'used_b',

    # User
    'acc_id', 'acc_fs_id', 'acc_logins', 'com_id', 'tz_id', 'com_alt_banner_height', 'com_alt_banner_width',
    'com_max_slots_per_res', 'com_min_hours_btw_res', 'active', 'logouts', 'acc_fs_id', 'acc_maint_notify',
    'acc_session_tx_number', 'acc_sys', 'acc_session_tx_number', 'accman_list_com_id', 'first_weekday',

    # VM
    'progress', 'vdc_id', 'vh_id', 'vh_cpu_cores', 'vh_cpu_count', 'vh_cpu_mhz', 'vh_cpu_threads', 'vh_memory_mb',
    'vh_pra_max_cpu', 'vh_pra_max_mem_mb', 'vh_pra_max_vm', 'vm_id', 'pc_id', 'pc_pod_id', 'pc_vm_id', 'vdc_id',
    'vh_id', 'vhg_id', 'vm_alloc_cpu_n', 'vm_alloc_mem_mb', 'vm_child_count', 'vm_parent_id', 'age_sec',

    # Paging
    'page_length', 'total_records', 'total_pages', 'page_limit', 'current_page', 'page_offset', 'page_total',
}

UUID_FIELDS = {
    # History
    'acc_uuid', 'cls_uuid', 'com_uuid', 'crs_uuid', 'lh_uuid', 'rh_uuid', 'lh_uuid_current', 'pod_uuid', 'res_uuid',
    'rh_uuid'
}

BLANK_IS_NONE_FIELDS = {
    # class
    'ros_team',
}


def deserialize(data):
    return json.loads(data, object_hook=process)


def resolve_enum(enum, value):
    if not isinstance(enum, tuple):
        enum = (enum,)
    for e in enum:
        try:
            return e(value)
        except:  # noqa
            pass
    raise ResponseFormatError("The value {} could not resolve to one of '{}'.".format(
        value, enum))


def process(obj):
    for field, value in obj.items():
        if field in DATE_FIELDS:
            obj[field] = to_date(value)
        if field in DATETIME_FIELDS:
            obj[field] = to_datetime(value)
        if field in TIMEDELTA_FIELDS:
            obj[field] = to_timedelta(value)
        if field in BOOLEAN_FIELDS:
            obj[field] = to_bool(value)
        if field in DECIMAL_FIELDS:
            obj[field] = to_decimal(value)
        if field in INT_FIELDS:
            obj[field] = to_int(value)
        if field in UUID_FIELDS:
            obj[field] = to_uuid(value)
        if field in ENUMS and value is not None:
            obj[field] = resolve_enum(ENUMS[field], value)
        if field in ENUM_CSV and value is not None:
            enum = ENUM_CSV[field]
            obj[field] = list(map(lambda v: enum[v], value.split(',')))
        if field in STR_CSV:
            if isinstance(value, str):
                obj[field] = value.split(',')
        if field in BLANK_IS_NONE_FIELDS and value == '':
            obj[field] = None
    return obj


def to_date(value):
    if value:
        return datetime.strptime(value, '%Y-%m-%d').date()
    else:
        return None


def to_datetime(value):
    if value:
        try:
            return datetime.strptime(value, '%Y-%m-%d %H:%M:%S')
        except ValueError:
            # rarely timestamps include microseconds. catch and process those
            return datetime.strptime(value, '%Y-%m-%d %H:%M:%S.%f')
    else:
        return None


def to_timedelta(value):
    if value:
        days, hours, minutes, seconds = [int(i) for i in value.split(',')]
        return timedelta(days=days, hours=hours, minutes=minutes, seconds=seconds)
    else:
        return None


def to_bool(value):
    truth_values = ('1', 'true', 'yes')
    normalized_value = str(value).lower()
    return normalized_value in truth_values


def to_decimal(value):
    if value:
        return Decimal(value)
    else:
        return None


def to_int(value):
    if value:
        return int(value)
    else:
        return 0


def to_uuid(value):
    if value:
        return UUID(value)
    else:
        return None


def serialize(data, pretty=False):
    extra_args: Any = {}
    if pretty:
        extra_args['indent'] = 2
        extra_args['sort_keys'] = True
    return json.dumps(data, default=default, **extra_args)


def default(obj):
    if isinstance(obj, Enum):
        return obj.value
    if isinstance(obj, datetime):
        return from_datetime(obj)
    if isinstance(obj, date):
        return from_date(obj)
    if isinstance(obj, timedelta):
        return from_timedelta(obj)
    if isinstance(obj, UUID):
        return from_uuid(obj)
    if isinstance(obj, Decimal):
        return from_decimal(obj)
    return obj


def from_date(value):
    return value.strftime("%Y-%m-%d")


def from_datetime(value):
    return value.strftime("%Y-%m-%d %H:%M:%S")


def from_timedelta(value):
    days, secs = value.days, value.seconds
    hours = secs // 3600
    minutes = (secs % 3600) // 60
    seconds = secs % 60
    return ','.join([str(days), str(hours), str(minutes), str(seconds)])


def from_uuid(value):
    return str(value)


def from_decimal(value):
    return str(value)



import asyncio
from datetime import datetime, timedelta

from netlab.async_client import NetlabClient
from netlab.enums import ReservationType
from netlab.errors.common import NetlabError
from netlab.errors.user import AccountNotFoundError

#####################
# Variables to change
#####################

POD_ID = 4402  # The pod to reserve
CLASS_ID = 2  # The class this reservation is for. Can only be set to None if the type below is INSTRUCTOR
RESERVATION_TYPE = ReservationType.ILT_CLASS  # choose from ('INDIVIDUAL'|'ILT_CLASS'|'TEAM'|'INSTRUCTOR')
RESERVATION_USERNAME = ''  # This is the user you want to make the reservation for.
EXERCISE_ID = ""

RESERVATION_TIME_ZONE = 'America/Chicago'  # Find your IANA timezone here - https://ndg.tech/hReahU
RESERVATION_START_TIME = '09:00'  # in HH:MM format, 24 hour time
RESERVATION_END_TIME = '12:00'  # in HH:MM format, 24 hour time
RESERVATION_FIRST_DAY = '2017-01-23'  # in YYYY-MM-DD format
RESERVATION_LAST_DAY = '2017-05-12'  # in YYYY-MM-DD format
# days of the week the reservations should be created, separated by spaces: M T W Th F Sa Su
RESERVATION_REPEAT_DAYS = 'M W F'  # This will repeat on Monday, Wednesday and Friday

NETLAB_SYSTEM = 'DEMO'  # must be a valid entry in your config.json

#########################
# Don't modify code below
#########################


async def main():
    days_of_the_week = ['M', 'T', 'W', 'Th', 'F', 'Sa', 'Su']
    repeat_days = RESERVATION_REPEAT_DAYS.split(" ")

    async with NetlabClient(NETLAB_SYSTEM) as client:
        reservations_created = 0
        reservations_with_errors = 0

        # find acc_id by looking up the username
        results = await client.user_account_search(
            properties=['acc_id'], filter={'acc_user_id': RESERVATION_USERNAME}, page=1)
        if results['data']:
            acc_id = results['data'][0]['acc_id']
        else:
            raise AccountNotFoundError(RESERVATION_USERNAME)

        # turn date strings into datetime objects
        first_day = datetime.strptime("{}".format(RESERVATION_FIRST_DAY), '%Y-%m-%d')
        last_day = datetime.strptime("{}".format(RESERVATION_LAST_DAY), '%Y-%m-%d')

        # loop through all the days between first and last day, inclusive
        current_day = first_day
        while current_day <= last_day:
            current_day_of_the_week = days_of_the_week[current_day.weekday()]
            if current_day_of_the_week in repeat_days:
                start_time = datetime.strptime('{} {}'.format(current_day.strftime('%Y-%m-%d'), RESERVATION_START_TIME),
                                               '%Y-%m-%d %H:%M')
                end_time = datetime.strptime('{} {}'.format(current_day.strftime('%Y-%m-%d'), RESERVATION_END_TIME),
                                             '%Y-%m-%d %H:%M')
                try:
                    await client.reservation_make(
                        type=RESERVATION_TYPE, pod_id=POD_ID, cls_id=CLASS_ID, start_time=start_time,
                        end_time=end_time, acc_id=acc_id, tz_olson=RESERVATION_TIME_ZONE,
                        ex_id=EXERCISE_ID)
                except NetlabError as err:
                    # The server returned an error. The exception will contain more information about the error.
                    reservations_with_errors += 1
                    print("  -! {}".format(err))
                else:
                    print("  -> Creating reservation at {}".format(start_time))
                    reservations_created += 1
            current_day += timedelta(days=1)

        print("\nSuccessfully created {} new reservations.".format(reservations_created))
        if reservations_with_errors > 0:
            print("There were {} errors encountered. Please review the output above.".format(reservations_with_errors))


if __name__ == '__main__':
    asyncio.run(main())



from .common import NetlabError, _add


@_add
class PodNotFoundError(NetlabError):
    """A pod could not be located from that request."""
    str_error = "E_POD_NOT_FOUND"


@_add
class DuplicatePodACLError(NetlabError):
    """Pod ACL already exists."""
    str_error = "E_DUPLICATE_POD_ACL"


@_add
class AdminAlreadyAuthorizedError(NetlabError):
    """acc_id is already authorized for pod_id."""
    str_error = "E_ADMIN_ALREADY_AUTHORIZED"


@_add
class ACLAdminNotFoundError(NetlabError):
    """Authorization for acc_id was not found for given pod_id."""
    str_error = "E_ACL_ADMIN_NOT_FOUND"


@_add
class PodACLNotFoundError(NetlabError):
    """pacl_uuid was not found."""
    str_error = "E_POD_ACL_NOT_FOUND"


@_add
class PodTypeNotFoundError(NetlabError):
    """pt_id was not found."""
    str_error = "E_POD_TYPE_NOT_FOUND"


@_add
class DuplicatePodIDError(NetlabError):
    """pod_id already exists."""
    str_error = "E_DUPLICATE_POD_ID"


@_add
class DuplicatePodNameError(NetlabError):
    """pod_name already exists."""
    str_error = "E_DUPLICATE_POD_NAME"


@_add
class PodMaxTypeReachedError(NetlabError):
    """Max of this pod type has been installed."""
    str_error = "E_POD_MAX_TYPE_REACHED"


@_add
class PodMaxLargeVLANReachedError(NetlabError):
    """Max number of pods with large VLANs (15) has been reached."""
    str_error = "E_POD_MAX_LARGE_VLAN_REACHED"


@_add
class PodMaxSysReachedError(NetlabError):
    """Max number of pods on this system has been reached."""
    str_error = "E_POD_MAX_SYS_REACHED"


@_add
class PodMaxDeviceReachedError(NetlabError):
    """Max number of devices has been reached."""
    str_error = "E_POD_MAX_DEVICE_REACHED"


@_add
class SourcePodNotFoundError(NetlabError):
    """The source_pod_id could not be found."""
    str_error = "E_SOURCE_POD_NOT_FOUND"


@_add
class LabDevicesNotSupportedError(NetlabError):
    """Pod cloning does not support real equipment."""
    str_error = "E_LAB_DEVICES_NOT_SUPPORTED"


@_add
class PodIDNotUniqueError(NetlabError):
    """The clone_pod_id is already in use."""
    str_error = "E_POD_ID_NOT_UNIQUE"


@_add
class PodNameNotUniqueError(NetlabError):
    """The clone_pod_name is already in use."""
    str_error = "E_POD_NAME_NOT_UNIQUE"


@_add
class NoPCCloneSpecError(NetlabError):
    """The pc_clone_spec is invalid."""
    str_error = "E_NO_PC_CLONE_SPEC"


@_add
class PCTypeNotSupportedError(NetlabError):
    """The pc_type in pc_clone_spec is not set to AVMI or NONE."""
    str_error = "E_PC_TYPE_NOT_SUPPORTED"


@_add
class SourceVirtualMachineIDRequired(NetlabError):
    """The source_vm_id in pc_clone_specs is required."""
    str_error = "E_SOURCE_VM_ID_REQUIRED"


@_add
class SourceVirtualMachineNotFoundError(NetlabError):
    """The source_vm_id in pc_clone_specs is not found."""
    str_error = "E_SOURCE_VM_NOT_FOUND"


@_add
class CloneTypeRequiredError(NetlabError):
    """The clone_type in pc_clone_specs is required."""
    str_error = "E_CLONE_TYPE_REQUIRED"


@_add
class SnapshotRequiredForLinkedCloneError(NetlabError):
    """The source_snapshot in pc_clone_specs was not set correctly."""
    str_error = "E_SNAPSHOT_REQUIRED_FOR_LINKED_CLONE"


@_add
class InvalidCloneTypeError(NetlabError):
    """The clone_type in pc_clone_specs must be set to FULL or LINKED."""
    str_error = "E_INVALID_CLONE_TYPE"


@_add
class CloneRoleRequiredError(NetlabError):
    """The clone_role in pc_clone_specs is required."""
    str_error = "E_CLONE_ROLE_REQUIRED"


@_add
class TemplateNotSupportedError(NetlabError):
    """clone_role in pc_clone_specs can not be set to TEMPLATE."""
    str_error = "E_TEMPLATE_NOT_SUPPORTED"


@_add
class VirtualMachineNameRequiredError(NetlabError):
    """The clone_name in pc_clone_specs is required."""
    str_error = "E_VM_NAME_REQUIRED"


@_add
class VirtualMachineNameInvalidError(NetlabError):
    """The clone_name in pc_clone_specs contains invalid characters."""
    str_error = "E_VM_NAME_INVALID"


@_add
class VirtualMachineNameNotUniqueLocalError(NetlabError):
    """The clone_name in pc_clone_specs is already in use."""
    str_error = "E_VM_NAME_NOT_UNIQUE_LOCAL"


@_add
class VirtualMachineNameNotUniqueGlobalError(NetlabError):
    """The clone_name in pc_clone_specs is already in use."""
    str_error = "E_VM_NAME_NOT_UNIQUE_GLOBAL"


@_add
class HostNotFoundError(NetlabError):
    """The clone_hostname in pc_clone_specs is not found."""
    str_error = "E_HOST_NOT_FOUND"


@_add
class HostOrGroupRequiredError(NetlabError):
    """The clone_vh_id in pc_clone_specs is required."""
    str_error = "E_HOST_OR_GROUP_REQUIRED"


@_add
class InvalidStorageAllocError(NetlabError):
    """The clone_storage_alloc in pc_clone_specs is not set to ONDEMAND or PREALLOCATED."""
    str_error = "E_INVALID_STORAGE_ALLOC"


@_add
class SameVirtualMachineHostRequiredError(NetlabError):
    """The source_pod_id pod type requires all VMs to be pointing to same clone_vh_id in pc_clone_specs."""
    str_error = "E_SAME_VM_HOST_REQUIRED"


@_add
class ClonePodNotFoundError(NetlabError):
    """The clone_pod_id is not found."""
    str_error = "E_CLONE_POD_NOT_FOUND"


@_add
class PodIDMismatchError(NetlabError):
    """Given pod_id does not match pc_pod_id."""
    str_error = "E_POD_ID_MISMATCH"


@_add
class PodIndexMismatchError(NetlabError):
    """Given pl_index does not match pc_pod_index."""
    str_error = "E_POD_INDEX_MISMATCH"


@_add
class VirtualMachineInUseError(NetlabError):
    """vm_id is currently in use."""
    str_error = "E_VM_IN_USE"


@_add
class PodNotOnlineError(NetlabError):
    """Current pod is not OFFLINE."""
    str_error = "E_POD_NOT_OFFLINE"


@_add
class VirtualMachineNotAttachedError(NetlabError):
    """vm_id not attached."""
    str_error = "E_VM_NOT_ATTACHED"


@_add
class PodInUseError(NetlabError):
    """Current pod is not OFFLINE."""
    str_error = "E_POD_IN_USE"


@_add
class PodActiveError(NetlabError):
    """The pod_id is not currently OFFLINE or ONLINE."""
    str_error = "E_POD_ACTIVE"


@_add
class PodNotSuspendedError(NetlabError):
    """Pod not in suspended state."""
    str_error = "E_POD_NOT_SUSPENDED"


@_add
class PodBusyError(NetlabError):
    """Pod is currently busy performing another task."""
    str_error = "E_POD_BUSY"


@_add
class PodEmptyError(NetlabError):
    """Pod has no devices or PCs."""
    str_error = "E_POD_EMPTY"


@_add
class PodNotActiveError(NetlabError):
    """Pod/reservation is not in active lab state."""
    str_error = "E_POD_NOT_ACTIVE"


@_add
class PodTypeIncompatibleError(NetlabError):
    """The pod type for the requested lab exercise is not the same as the reserved pod type."""
    str_error = "E_POD_TYPE_INCOMPATIBLE"



from enum import Enum


class FreeAlarm(Enum):
    "These values are triggered when a disk's free space has dropped below a determined threshold."
    CLEAR = 'CLEAR'
    "The disk has adequate free space."
    WARN = 'WARN'
    "The disk has reached free space warning"
    MAJOR = 'MAJOR'
    "The disk is nearly out of free space."
    CRITICAL = 'CRITICAL'
    "The disk is out of free space."



"""
Netlab enums.
"""

from .common import DateFormat, HDRSeverity, HDRStatus, TimeFormat
from .klass import ClassEmailLogs, ClassExtensionSlots, ClassLabLimit, ContentAccessiblity
from .pod import (
    PCIcon,
    PCType,
    PLIcon,
    PodAdminState,
    PodCategory,
    PodCurrentState,
    PodState,
    RemoveVMS,
    VirtualHostComPath,
    VirtualMachineRole,
)
from .reservation import InitConfig, ReservationScope, ReservationType
from .system import FreeAlarm
from .user import AccountPrivileges, AccountType, CommunityExtensionSlots
from .vm import CloneStorageAllocation, CloneType, PowerState, VirtualMachineLicenseType

__all__ = (
    'TimeFormat', 'DateFormat', 'HDRSeverity', 'HDRStatus',

    'ClassEmailLogs', 'ClassLabLimit', 'ClassExtensionSlots',

    'PCIcon', 'PCType', 'PLIcon', 'PodCategory', 'PodState', 'RemoveVMS', 'VirtualHostComPath',
    'VirtualMachineRole', 'PodAdminState', 'PodCurrentState',

    'ReservationType', 'ReservationScope', 'InitConfig',

    'FreeAlarm',

    'AccountPrivileges', 'AccountType', 'CommunityExtensionSlots', 'ContentAccessiblity',

    'PowerState', 'CloneStorageAllocation', 'CloneType', 'VirtualMachineLicenseType'
)



"""
Netlab SDK api methods
"""

from .klass import ClassApiMixin
from .lab import LabApiMixin
from .pod import PodApiMixin
from .reservation import ReservationApiMixin
from .system import SystemApiMixin
from .user import UserApiMixin
from .vm import VmApiMixin

__all__ = [
    'ClassApiMixin', 'PodApiMixin', 'ReservationApiMixin', 'SystemApiMixin', 'LabApiMixin', 'UserApiMixin',
    'VmApiMixin'
]



import asyncio

from netlab import enums
from netlab.async_client import NetlabClient, NetlabConnection

SOURCE_POD_ID = 9000
POD_NAME_TEMPLATE = 'example_pod_{}'
POD_RANGE = range(900, 910)


async def clone_pod(client: NetlabConnection, pod_number: int):
    pod_name = POD_NAME_TEMPLATE.format(pod_number)

    await client.pod_clone_task(
        source_pod_id=SOURCE_POD_ID,
        clone_pod_id=pod_number,
        clone_pod_name=pod_name,
        pc_clone_specs=[{
            "source_snapshot": "master",
            "clone_snapshot": "master",
        }])
    await client.pod_state_change(pod_id=pod_number, state=enums.PodState.ONLINE)

    return pod_name


async def main():
    async with NetlabClient() as client:
        for pod_task in asyncio.as_completed([
            clone_pod(client, i) for i in POD_RANGE
        ]):
            pod_name = await pod_task

            print(f'Cloned pod: {pod_name}')


if __name__ == '__main__':
    asyncio.run(main())



import datetime

from netlab.enums import PodState
from netlab.sync_client import SyncClient

"""
The intent of this code is to help you, the NETLAB+ administrator, learn how you can use the
NETLAB+ API to automate repetitive and tasks associate with administering NETLAB+. This sample
file contains functions that identify how to make an API call and what you can do with it's
output.

You will need to have a running version of python 3 (python 3.5 or greater is best) and install pip.
The netlab api can be obtained by running the following PIP command:

     pip install https://ndg.tech/netlab-py-latest

You are encouraged to use code in this document as you see fit for your environment.

If you are new to programming, do not fear. You can copy this document or functions in the document,
alter noted variables and be up and running in minutes.
"""

# The system will need to be changed to the name given to your system in the .netlab/config.json file
# or create a entry for DEMO in your config.json.
tapi = SyncClient(system='DEMO')


def pod_add():
    """
    This function will add pods with corresponding pod designs from pods listed below.

    The pod type id (pt_id) is the pod design UUID and is subject to change with new revisions of the
    pod design or new course.

    Also note that the the pod id (pod_id) and pod name (pod_name) does not need to follow the scheme
    as they are layed out below. They can be adjusted to your preference.
    """
    # tapi.pod_add(pod_id=1000, pt_id=str("RHSA7_0050_56A9_38CC_5682_A22E"), pod_name="RHSA7_GM")
    # tapi.pod_add(pod_id=2000, pt_id=str("NDGEH_0050_56A9_38CC_5612_BA75"), pod_name="NDG_EH_GM")
    # tapi.pod_add(pod_id=3000, pt_id=str("NDGDF_0050_56A9_38CC_568E_877A"), pod_name="NDG_Forensics_GM")
    # tapi.pod_add(pod_id=4999, pt_id=str("PAN7GW_0050_56A9_38CC_5682_D5F0"), pod_name="PAN7_GW_GM")
    # tapi.pod_add(pod_id=4000, pt_id=str("PAN7FE_0050_56A9_38CC_5682_D3B7"), pod_name="PAN7_FE_GM")
    # tapi.pod_add(pod_id=5000, pt_id=str("VMVCADCV6_0050_56A9_38CC_5628_F6DE"), pod_name="VM_VCA_DCV_6_GM")
    # tapi.pod_add(pod_id=6000, pt_id=str("VCPICM6F1_0050_56A9_DC6B_5643_78BE"), pod_name="VM_VCP_ICM_6_GM")
    # tapi.pod_add(pod_id=7000, pt_id=str("VMWOS6_0050_56A9_38CC_5630_CB6A"), pod_name="VM_VCP_ONS_6_GM")
    # tapi.pod_add(pod_id=8000, pt_id=str("EMCCIS01_0050_56B3_0CC0_54C0_03A9"), pod_name="EMC_CIS_1_GM")
    # tapi.pod_add(pod_id=9000, pt_id=str("EMCISMV2_0050_56B3_0CC0_54BF_D4A1"), pod_name="EMC_ISM_v2_GM")
    # tapi.pod_add(pod_id=10000, pt_id=str("DOLPYSEC_0050_56A9_38CC_54B4_797A"), pod_name="NISGTC_PySec_GM")
    # tapi.pod_add(pod_id=11000, pt_id=str("DOLAPV2_0050_56A9_3CF1_54BE_CE1D"), pod_name="NISGTC_APlus_v2_GM")
    # tapi.pod_add(pod_id=12000, pt_id=str("DOLLPINST_0025_9015_B226_50F7_F3AF"), pod_name="NISGTC_LPlus_Install_GM")
    # tapi.pod_add(pod_id=13000, pt_id=str("DOLLP_0025_9015_B226_50B5_47A2"), pod_name="NISGTC_LPlus_Base_GM")
    # tapi.pod_add(pod_id=14000, pt_id=str("DOLNETP_0025_9015_B226_519E_15D6"), pod_name="NISGTC_NetPlus_GM")
    # tapi.pod_add(pod_id=15000, pt_id=str("DOLNSEC_0050_56A9_38CC_53C5_7988"), pod_name="NISGTC_NetSec_GM")
    # tapi.pod_add(pod_id=16000, pt_id=str("DOLSECPV2_0050_56A9_38CC_54C1_5A23"), pod_name="NISGTC_SecPlus_GM")


def pod_clone(src_pid=None, clone_pid_rng=None, clone_pname=None, pc_clone_specs=None):
    """
    Clones an existing pod.

    The pod_clone_task() available via the API is one of the most complex, requested and used of the methods
    available in the API. There are many variables that can be changed to allow cloning of pods to changing
    hardware configurations as well as various pod configurations.

    Required information includes:
    source_pod_id: the numerical pod identifier of the pod you wish to clone, normally this is the Master.
    clone_pod_id: the numerical pod id that you will assign to the pod.
    clone_pod_name: the name given to the cloned instance of the pod.
        - <Program><Course><DataStore><Host><PodId>
          NDG_EH_L_H81_P1001

    Alternate information predominantly include data in the pc_clone_specs that you may need to change for
    operations in your environment. This will normally specify on which host and on which datastore the
    virtual machine will reside.

    The following shows how to specify the pc_clone_specs which sets the datastore "datastore1" on ESXi
    host id 1. To find the vh_id of a host, see the docs for the API method `vm_host_find`.

    pc_clone_specs=[{"clone_datastore": "datastore1", "clone_vh_id": 1}]

    """
    # TODO: JJ/TK: need to ensure that we give a best practice of for enumerating pods and pod naming conventions.
    for pid in clone_pid_rng:
        print(datetime.datetime.now())
        output = tapi.pod_clone_task(source_pod_id=src_pid, clone_pod_id=pid, clone_pod_name=clone_pname+str(pid),
                                     pc_clone_specs=pc_clone_specs)
        print("Cloned Pod: " + str(pid) + '\t' + str(output['status']))


def pod_list():
    """
    Retrievs a list of pods on your NETLAB+ VE system.

    Initially, this method in the API will return
    """
    # Initiate the nlapi.pod_lst() method to bring back a list of dictionaries of key/value pirs of data.

    mlist = tapi.pod_list()

    # If you want an easy way to see your returned values and have values appear in columns you can
    # use the following for loop to iterate through the returned records and print their values:
    for i in mlist:
        output = '{0}\t{1}\t{2}\t{3}\n'.format(str(i['pod_id']), str(i['pod_name']), str(i['pod_desc']),
                                               str(i['pod_uuid']))
        print(output)

    # This will return the contents of the mlist variable to the calling function.
    return mlist


def pod_get(pod_id=int()):
    """
    Retrieves information about an individual pod. Requires pod_id for input.
    """
    result = tapi.pod_get(pod_id=pod_id, properties=['pod_id', 'pod_name', 'pod_desc', 'pod_uuid'])
    print(result)

    return result


def pod_remove(pidrange, remove_vms="DISK"):
    """
    Removes an existing pod from the NETLAB+ pod inventory.

    This is comparable to selecting a pod from the NETLAB+ pod list and clicking "Delete".

    As with the web interface, you can what you do with the virtual machines when you
    execute. The following is the list of valid parameters: 'NONE', 'LOCAL', 'DATACENTER', 'DISK'.

    - 'NONE':  Do not delete any VMs (they will remain in the NETLAB+ inventory)
    - 'LOCAL':  Remove VMs from NETLAB+ inventory only (VMs remains in datacenter)
    - 'DATACENTER': Remove VMs from NETLAB+ inventory and datacenter (VM files not deleted from disk)
    - 'DISK': Remove VMs from NETLAB+ inventory, datacenter, AND delete unshared VM files from disk

    nlapi.pod_remove_task(pod_id=13001, remove_vms="DISK")

    """
    # Check to see if if the pidrange variable is range or int type
    if isinstance(pidrange, range):
        for pid in pidrange:
            # First, set or ensure the pod is set to offline
            tapi.pod_state_change(pod_id=pid, state=PodState.OFFLINE)
            offline_time = datetime.datetime.now()
            print("Pod Offline: " + str(pid) + '\tOK\t' + str(offline_time))
            # Now we can remove the pod.
            result = tapi.pod_remove_task(pod_id=pid, remove_vms=remove_vms)
            removed_time = datetime.datetime.now()
            print("Pod Removed: " + str(pid) + '\t' + str(result['status']) + "\t" + str(removed_time))
    elif isinstance(pidrange, int):
        tapi.pod_state_change(pod_id=pidrange, state=PodState.OFFLINE)
        offline_time = datetime.datetime.now()
        print("Pod Offline: " + str(pidrange) + '\tOK\t' + str(offline_time))
        # Now we can remove the pod.
        result = tapi.pod_remove_task(pod_id=pidrange, remove_vms=remove_vms)
        removed_time = datetime.datetime.now()
        print("Pod Removed: " + str(pidrange) + '\t' + str(result['status']) + "\t" + str(removed_time))


def pod_online(pidrange):
    """
    Brings pod or pods online.

    This function accepts either an integer for the pod_id or a range object.
    You can specify a range of pods that you wish to loop through or an individul pod id.

    pod_online(1001) or pod_online(range(1001, 1011))

    Uses the pod_state_changed() method in the NETLAB+ API.
    """
    if isinstance(pidrange, range):
        for pid in pidrange:
            tapi.pod_state_change(pod_id=pid, state=PodState.OFFLINE)
            print("Brought Online: " + str(pid) + "\tOK")
    elif isinstance(pidrange, int):
        tapi.pod_state_change(pod_id=pidrange, state=PodState.OFFLINE)
        print("Brought Online: " + str(pidrange) + "\tOK")


def pod_offline(pidrange):
    """
    Takes pod or pods offline.

    This function accepts either an integer for the pod_id or a range object.
    You can specify a range of pods that you wish to loop through or an individul pod id.

    pod_offline(1001) or pod_offline(range(1001, 1011))

    Uses the pod_state_changed() method in the NETLAB+ API.
    """
    if isinstance(pidrange, range):
        for pid in pidrange:
            tapi.pod_state_change(pod_id=pid, state=PodState.OFFLINE)
            print("Brought Offline: " + str(pid) + "\tOK")
    elif isinstance(pidrange, int):
        tapi.pod_state_change(pod_id=pidrange, state=PodState.OFFLINE)
        print("Brought Offline: " + str(pidrange) + "\tOK")


def user_list(search_param=''):
    """
    List users and restrict the output to a passed search parameter.

    user_list('student1')
    """
    temp_list = tapi.user_account_list()
    del_acc_ids = []
    for x in temp_list:
        if x['acc_user_id'].startswith(search_param):
            ret_acc_id = x['acc_id']
            ret_acc_user_id = x['acc_user_id']
            print('acc_id=' + str(ret_acc_id) + '\tacc_user_id=' + str(ret_acc_user_id))
            del_acc_ids.append(ret_acc_id)
    return del_acc_ids


def user_add(com_id, acc_user_id, acc_password, acc_full_name, acc_type="S", cls_id=None):
    """
    Add a user to your system.

    You can add other parameters to the user_account_add() method.

    Other parameters  that may be of interest:

    - cls_id : Class ID will allow you to add the user to a class when creating the user.
    - tz_id : Time Zone ID allows you set the set the time zone for the user upon creation.
    - acc_type: Account Type, "S" for student and "I" for instructor accounts.

    user_add(com_id, acc_user_id, acc_password, acc_full_name, acc_type="S",
            cls_id, tz_id, acc_display_name, acc_sort_name, acc_email )
    """

    tapi.user_account_add(com_id=com_id, acc_user_id=acc_user_id, acc_password=acc_password,
                          acc_full_name=acc_full_name, acc_type=acc_type, cls_id=cls_id)
    return


def user_del(search_param=''):
    """
    Delete a user from the system.

    user_del('student1')
    """
    del_list = user_list(search_param=search_param)
    for x in del_list:
        tapi.user_account_remove(acc_id=x)
        print('acc_id=' + str(x) + '\tdeleted')
    return


def class_add(cls_name, com_id):
    """
    Adds a class to system.

    class_add(cls_name, com_id)

    cls_name: The name of the class
    com_id: The community ID that the class belongs to.

    """
    # TODO: JJ: What if com_id is not known? - Can go to web gui and obtain?
    result = tapi.class_add(cls_name=cls_name, com_id=com_id)
    print(result)
    return result


def class_remove(cls_id, delete_students=False):
    """
    Removes a class from system.

    class_remove(cls_id, delete_students=False)

    cls_id: Class ID - can be obtained via class_list() method.
    """
    tapi.class_remove(cls_id=cls_id, delete_students=delete_students)


def class_get(cls_id):
    """
    Gets information about class.


    """
    result = tapi.class_get(cls_id=cls_id)
    print(result)
    return result


def class_list(com_id='1'):
    """
    List available classes on system.


    """
    result = tapi.class_list(com_id=com_id)
    for iklass in result:
        print('output: \t' + str(iklass['cls_id']) + '\t' + iklass['cls_name'] + '\t' + iklass['cls_uuid'] + '\n')
    return


def lplus_base():
    # Specify the pod id of the master pod.
    pod_master = 13000

    # pod_rng specifies a linear range of pod id's that you wish to manipulate.
    pod_rng = range(13006, 13008)

    # Set a prefix for pod identification.
    # A good prefix will include identifiers for course/lab set, datastore location and host. In the sample
    # below we also supply a P at the end as the pod_id will be attached to it, please see pod_clone().
    pod_prefix = 'LPlus_Base_L_H81_P'

    pod_clone(src_pid=pod_master, clone_pid_rng=pod_rng, clone_pname=pod_prefix,
              pc_clone_specs={"clone_datastore": "MCNC_HOST81_LOCAL", "clone_vh_id": 1})
    # Uncomment the following if you wish to bring pods online. You can change the pod_online()
    # function to pod_offline() to take the pods offline.
    # for i in pod_rng:
    #     pod_online(i)
    # Uncomment the following line to remove the pods in the pod id range (pod_rng) passed to the function.
    # pod_remove(pod_rng)



import warnings
from typing import Any, Dict, List, Literal, Optional, overload

from .. import enums
from ..errors.vm import VirtualMachineDatacenterNotFoundError
from ..utils import minimum_version
from ._client_protocol import ClientProtocol


class VmApiMixin(ClientProtocol):
    async def vm_datacenter_get(
                self,
                *,
                vdc_id: int,
                **kwargs
            ) -> Dict[str, Any]:
        """
        This method allows you to retrieve a datacenter's information by vdc_id.

        :param vdc_id: VM Datacenter identifier.

        :return: On success, returns an object with the following properties:

        **Properties**

        vdc_agent_hostname
            Agent hostname/IP address.
        vdc_agent_password
            Agent password.
        vdc_agent_username
            Agent username.
        vdc_date_added
            Timestamp of when VDC was added to NETLAB+.
        vdc_date_tested
            Timestamp of when VDC was last tested.
        vdc_last_test_status
            Status of last VDC test.
        vdc_id
            VM Datacenter identifier.
        vdc_name
            VM Datacenter name.
        vdt_type
            VM Datacenter type.
        vdt_name
            VM Datacenter type name.
        """

        method = "vm.datacenter.get"
        return await self.call(method, vdc_id=vdc_id, **kwargs)

    async def vm_datacenter_find(
                self,
                *,
                vdc_name: str,
                **kwargs
            ) -> Optional[int]:
        """
        This method allows you to find a datacenter's **vdc_id** by the datacenter name.

        :param vdc_name: VM Datacenter name.

        :return: On success, returns vdc_id. If there is no VDC matching, returns None.
        """

        method = "vm.datacenter.get"
        try:
            result = await self.call(method, vdc_name=vdc_name, **kwargs)
            return result['vdc_id']
        except VirtualMachineDatacenterNotFoundError:
            return None

    async def vm_datacenter_list(self, **kwargs) -> List[Dict[str, Any]]:
        """
        This method allows you to retrieve all datacenters' information.

        :return: A list of objects with the following properties:

        **Properties**

        vdc_agent_hostname
            Agent hostname/IP address.
        vdc_agent_password
            Agent password.
        vdc_agent_username
            Agent username.
        vdc_date_added
            Timestamp of when VDC was added to NETLAB+.
        vdc_date_tested
            Timestamp of when VDC was last tested.
        vdc_last_test_status
            Status of last VDC test.
        vdc_id
            VM Datacenter identifier.
        vdc_name
            VM Datacenter name.
        vdt_type
            VM Datacenter type.
        vdt_name
            VM Datacenter type name.
        """

        method = "vm.datacenter.list"
        return await self.call(method, **kwargs)

    async def vm_datacenter_test(
                self,
                *,
                vdc_id: int,
                **kwargs
            ) -> Literal['PASSED', 'FAILED']:
        """
        This method allows you to call a task to test communication to datacenter.

        :param vdc_id: VM Datacenter identifier.

        :return: On success returns 'PASSED', otherwise returns 'FAILED'.
        """

        method = "vm.datacenter.test.task"
        return await self.call(method, vdc_id=vdc_id, **kwargs)

    async def vm_host_get(
                self,
                *,
                vh_id: int,
                **kwargs
            ) -> Dict[str, Any]:
        """
        This method allows you to retrieve a virtual host information.

        :param vh_id: Virtual host identifier.

        :return: On success, returns an object with the following properties:

        **Properties**

        vdc_id
            Agent hostname/IP address.
        vdc_name
            VM Datacenter name.
        vdt_type
            VM Datacenter type.
        vdt_name
            VM Datacenter type name.
        vh_bios_date
            Timestamp of BIOS creation date.
        vh_bios_version
            BIOS version installed.
        vh_com_path
            Virtual host communication path.
        vh_cpu_cores
            Total count of CPU cores on virtual host.
        vh_cpu_count
            Total count of physical CPUs on virtual host.
        vh_cpu_mhz
            Speed in MHz of CPU.
        vh_cpu_model
            CPU model identification.
        vh_cpu_threads
            Number of threads CPUs can handle.
        vh_date_added
            Timestamp of virtual host added to NETLAB+.
        vh_date_modified
            Timestamp of virtual host information has been modified.
        vh_date_tested
            Timestamp of virtual host test communication.
        vh_id
            Virtual host identifier.
        vh_inside_ipv4_addr
            Inside interface ipv4 address.
        vh_inside_vswitch_0
            Virtual host inside vSwitch.
        vh_last_test_status
            Status of last virtual host test communication.
        vh_memory_mb
            Total RAM in MB of virtual host.
        vh_name
            Virtual host hostname.
        vh_online
            Virtual host is online.
        vh_os_build
            Operating system build number.
        vh_os_description
            Operating system description.
        vh_os_name
            Operating system name.
        vh_os_type
            Operating system type.
        vh_os_vendor
            Operating system vender.
        vh_os_version
            Operating system version.
        vh_outside_ipv4_addr
            Outside interface ipv4 address.
        vh_pra_enabled
            Proactive resource awareness is enabled.
        vh_pra_max_cpu
            Max number of CPUs for PRA.
        vh_pra_max_mem_mb
            Max amount of RAM in MB for PRA.
        vh_pra_max_vm
            Max amount of virtual machines for PRA.
        vh_sys_model
            System model.
        vh_sys_service_tag
            System service tag.
        vh_sys_vendor
            System vendor.
        vh_uuid
            Virtual host unique global identifier.
        """

        method = "vm.host.get"
        return await self.call(method, vh_id=vh_id, **kwargs)

    async def vm_host_find(
                self,
                *,
                vh_name: str,
                **kwargs
            ) -> Optional[int]:
        """
        This method allows you to find a virtual host's **vh_id** by the virtual host name.

        :param vh_name: Virtual Host name.

        :return: On success, returns vh_id. If there is no matching Virtual Host, returns None.
        """

        method = "vm.host.list"

        hosts = await self.call(method, **kwargs)

        for host in hosts:
            if host.get('vh_name') == vh_name:
                return host['vh_id']
        else:
            return None

    async def vm_host_list(
                self,
                *,
                vdc_id: Optional[str] = None,
                **kwargs
            ) -> List[Dict[str, Any]]:
        """
        This method allows you to retrieve a list of virtual hosts information.

        :param vdc_id: VM Datacenter identifier. If not specified, will return from all datacenters.

        :return: Returns a list of objects with the following properties:

        **Properties**

        vdc_id
            Agent hostname/IP address.
        vdc_name
            VM Datacenter name.
        vdt_type
            VM Datacenter type.
        vdt_name
            VM Datacenter type name.
        vh_bios_date
            Timestamp of BIOS creation date.
        vh_bios_version
            BIOS version installed.
        vh_com_path
            Virtual host communication path.
        vh_cpu_cores
            Total count of CPU cores on virtual host.
        vh_cpu_count
            Total count of physical CPUs on virtual host.
        vh_cpu_mhz
            Speed in MHz of CPU.
        vh_cpu_model
            CPU model identification.
        vh_cpu_threads
            Number of threads CPUs can handle.
        vh_date_added
            Timestamp of virtual host added to NETLAB+.
        vh_date_modified
            Timestamp of virtual host information has been modified.
        vh_date_tested
            Timestamp of virtual host test communication.
        vh_id
            Virtual host identifier.
        vh_inside_ipv4_addr
            Inside interface ipv4 address.
        vh_inside_vswitch_0
            Virtual host inside vSwitch.
        vh_last_test_status
            Status of last virtual host test communication.
        vh_memory_mb
            Total RAM in MB of virtual host.
        vh_name
            Virtual host hostname.
        vh_online
            Virtual host is online.
        vh_os_build
            Operating system build number.
        vh_os_description
            Operating system description.
        vh_os_name
            Operating system name.
        vh_os_type
            Operating system type.
        vh_os_vendor
            Operating system vender.
        vh_os_version
            Operating system version.
        vh_outside_ipv4_addr
            Outside interface ipv4 address.
        vh_pra_enabled
            Proactive resource awareness is enabled.
        vh_pra_max_cpu
            Max number of CPUs for PRA.
        vh_pra_max_mem_mb
            Max amount of RAM in MB for PRA.
        vh_pra_max_vm
            Max amount of virtual machines for PRA.
        vh_sys_model
            System model.
        vh_sys_service_tag
            System service tag.
        vh_sys_vendor
            System vendor.
        vh_uuid
            Virtual host unique global identifier.
        """

        method = "vm.host.list"
        return await self.call(method, vdc_id=vdc_id, **kwargs)

    async def vm_inventory_add(
                self,
                *,
                vdc_id,
                vm_uuid,
                vm_name,
                vm_path=None,
                vm_alloc_mem_mb,
                vm_alloc_cpu_n,
                vhg_id=None,
                vh_id=None,
                vm_role=enums.VirtualMachineRole.NORMAL,
                vm_comments=None,
                vm_vendor_os_id=None,
                vm_vendor_os_name=None,
                vm_netlab_os_id=None,
                vm_parent_id=None,
                vm_parent_snapname=None,
                **kwargs
            ) -> List[Dict[str, Any]]:
        """
        *``vm_inventory_add`` is deprecated, please use ``vm_inventory_import_task`` instead*

        Add virtual machines to NETLAB+ inventory.

        :param int vdc_id: VM Datacenter identifier.
        :param int vhg_id: Virtual host group identifier.
        :param int vh_id: Virtual host identifier.
        :param str vm_uuid: Virtual machine unique identifier.
        :param str vm_name: Virtual machine name.
        :param str vm_path: Virtual machine path on datastore.
        :param netlab.enums.VirtualMachineRole vm_role: Virtual machine role in NETLAB+ inventory.
        :param int vm_alloc_mem_mb: Total amount of allocated RAM.
        :param int vm_alloc_cpu_n: Total amount of allocated CPUs.
        :param str vm_comments: Virtual machine comments from VMX.
        :param str vm_vendor_os_id: Vendor operating system identifier.
        :param str vm_vendor_os_name: Vendor operating system name.
        :param str vm_netlab_os_id: NETLAB+ match to vm_vendor_os_id.
        :param int vm_parent_id: Virtual machine parent identifier, if VM is a link clone.
        :param str vm_parent_snapname: Virtual machine parent snapname, if VM is a link clone.

        :return: Returns a list of object with the following properties:
        :rtype: list[dict]

        **properties**

        vm_id
            Virtual machine identifier.
        """
        warnings.warn("'vm_inventory_add' is deprecated, please use 'vm_inventory_import_task' instead",
                      DeprecationWarning)

        method = 'vm.inventory.add'
        return await self.call(
            method, vdc_id=vdc_id, vm_uuid=vm_uuid, vm_name=vm_name, vm_path=vm_path,
            vm_alloc_mem_mb=vm_alloc_mem_mb, vm_alloc_cpu_n=vm_alloc_cpu_n, vhg_id=vhg_id, vh_id=vh_id,
            vm_role=vm_role, vm_comments=vm_comments, vm_vendor_os_id=vm_vendor_os_id,
            vm_vendor_os_name=vm_vendor_os_name, vm_netlab_os_id=vm_netlab_os_id,
            vm_parent_id=vm_parent_id, vm_parent_snapname=vm_parent_snapname, **kwargs)

    async def vm_inventory_import_task(
                self,
                *,
                vdc_id: int,
                vm_property_list: List[Dict[str, Any]],
                **kwargs
            ) -> List[Dict[str, Any]]:
        """
        Method to call a task to import virtual machine into NETLAB+ inventory.

        :param int vdc_id: Virtual datacenter identifier.
        :param list vm_property_list: Virtual machine property list of the following properties:

        **properties**

        vm_uuid
            Virtual machine unique identifier.
        vhg_id
            Virtual host group identifier.
        vh_id
            Virtual host identifier.
        vm_name
            Virtual machine name.
        vm_role
            Virtual machine role in NETLAB+ inventory. :py:class:`netlab.enums.VirtualMachineRole`
        vm_alloc_mem_mb
            Total amount of allocated RAM.
        vm_alloc_cpu_n
            Total amount of allocated CPUs.
        vm_comments
            Virtual machine comments from VMX.
        vm_vendor_os_id
            Vendor operating system identifier.
        vm_vendor_os_name
            Vendor operating system name.
        vm_netlab_os_id
            NETLAB+ match to `vm_vendor_os_id`.

        :return: List of results

        **return properties**

        status
            'OK' or error.
        vm_id
            Virtual machine identifier.
        """
        method = 'vm.inventory.import.task'
        return await self.call(method, vdc_id=vdc_id, vm_property_list=vm_property_list, **kwargs)

    async def vm_inventory_get(
                self,
                *,
                vm_id: int,
                **kwargs
            ) -> Dict[str, Any]:
        """
        This method allows you to retrieve one or more virtual machines from NETLAB+ inventory.

        :param vm_id: Virtual machine identifier.

        :return: Returns an object with the following properties:

        **Properties**

        pc_id
            Remote pc identifier.
        pc_label
            Remote pc label.
        pc_os_id
            NETLAB+ operating system identifier.
        pc_os_name
            NETLAB+ operating system name.
        pc_pod_id
            If attached to a pod, pod identifier.
        pc_vm_id
            Virtual machine identifier.
        pod_name
            Pod name.
        vdc_id
            Virtual datacenter identifier.
        vdc_name
            Virtual datacenter name.
        vh_id
            Virtual host identifier.
        vh_name
            Virtual host name.
        vhg_id
            Virtual host group identifier.
        vhg_name
            Virtual host group name.
        vm_alloc_cpu_n
            Total amount of allocated CPUs.
        vm_alloc_mem_mb
            Total amount of allocated RAM.
        vm_child_count
            Total count of children VMs.
        vm_comments
            Virtual machine comments from VMX.
        vm_date_added
            Timestamp of virtual machine added to NETLAB+ inventory.
        vm_id
            Virtual machine identifier.
        vm_name
            Virtual machine name.
        vm_netlab_os_id
            NETLAB+ match to vm_vendor_os_id.
        vm_parent_id
            Virtual machine parent identifier, if VM is a link clone.
        vm_parent_name
            Virtual machine parent name.
        vm_parent_role
            Parent virtual machine role.
        vm_parent_snapname
            Virtual machine parent snapname, if VM is a link clone.
        vm_path
            Virtual machine path on datastore.
        vm_power_state
            Virtual machine power state.
        vm_role
            Virtual machine role in NETLAB+ inventory.
        vm_snapshot
            This is what the VM reverts to when the pod is reset. You can take multiple snapshots with the same name,
            without having to update this field. Update this field with a call to
            :meth:`~netlab.api.PodApiMixin.pod_pc_update`
        vm_uuid
            Virtual machine unique identifier.
        vm_vendor_os_id
            Vendor operating system identifier.
        vm_vendor_os_name
            Vendor operating system name.

        """
        method = "vm.inventory.get"
        return await self.call(method, vm_id=vm_id, **kwargs)

    async def vm_inventory_list(
                self,
                *,
                vdc_id: Optional[int] = None,
                roles: Optional[List[str]] = None,
                attached: Optional[bool] = None,
                **kwargs
            ) -> List[Dict[str, Any]]:
        """
        This method allows you to list the virtual machines from NETLAB+ inventory.

        :param vdc_id: VM Datacenter identifier.
        :param roles: Filter by specified roles.
        :param attached: List virtual machines attached to a pod.


        :return: Returns an object with the following properties:

        **Properties**

        pc_id
            Remote pc identifier.
        pc_label
            Remote pc label.
        pc_os_id
            NETLAB+ operating system identifier.
        pc_os_name
            NETLAB+ operating system name.
        pc_pod_id
            If attached to a pod, pod identifier.
        pc_vm_id
            Virtual machine identifier.
        pod_name
            Pod name.
        vdc_id
            Virtual datacenter identifier.
        vdc_name
            Virtual datacenter name.
        vh_id
            Virtual host identifier.
        vh_name
            Virtual host name.
        vhg_id
            Virtual host group identifier.
        vhg_name
            Virtual host group name.
        vm_alloc_cpu_n
            Total amount of allocated CPUs.
        vm_alloc_mem_mb
            Total amount of allocated RAM.
        vm_child_count
            Total count of children VMs.
        vm_comments
            Virtual machine comments from VMX.
        vm_date_added
            Timestamp of virtual machine added to NETLAB+ inventory.
        vm_id
            Virtual machine identifier.
        vm_name
            Virtual machine name.
        vm_netlab_os_id
            NETLAB+ match to vm_vendor_os_id.
        vm_parent_id
            Virtual machine parent identifier, if VM is a link clone.
        vm_parent_name
            Virtual machine parent name.
        vm_parent_role
            Parent virtual machine role.
        vm_parent_snapname
            Virtual machine parent snapname, if VM is a link clone.
        vm_path
            Virtual machine path on datastore.
        vm_power_state
            Virtual machine power state.
        vm_role
            Virtual machine role in NETLAB+ inventory.
        vm_snapshot
            This is what the VM reverts to when the pod is reset. You can take multiple snapshots with the same name,
            without having to update this field. Update this field with a call to
            :meth:`~netlab.api.PodApiMixin.pod_pc_update`
        vm_uuid
            Virtual machine unique identifier.
        vm_vendor_os_id
            Vendor operating system identifier.
        vm_vendor_os_name
            Vendor operating system name.
        """
        method = "vm.inventory.get"
        params: Dict[str, Any] = {}

        if vdc_id:
            params['vdc_id'] = vdc_id
        if roles:
            params['roles'] = roles
        if attached is not None:
            params['attached'] = attached

        params.update(kwargs)
        return await self.call(method, **params)

    async def vm_snapshot_add(
                self,
                *,
                vm_id: Optional[int] = None,
                pc_id: Optional[int] = None,
                snapshot_name: str,
                dump_memory: bool = False,
                sync: bool = False,
                description: Optional[str] = None,
                **kwargs
            ) -> None:
        """
        Method to call a task to add snapshot to virtual machine.

        :param vm_id: Virtual datacenter identifier.
        :param pc_id: Remote pc identifier.
        :param snapshot_name: Name of snapshot to take.
        :param dump_memory: Dump memory files.
        :param sync: Sync files.
        :param description: Snapshot description.
        """

        method = "vm.snapshot.add.task"
        return await self.call(
            method, vm_id=vm_id, pc_id=pc_id, snapshot_name=snapshot_name, dump_memory=dump_memory,
            sync=sync, description=description, **kwargs)

    async def vm_snapshot_edit(
                self,
                *,
                vm_id: Optional[int] = None,
                pc_id: Optional[int] = None,
                snapshot_id: Optional[int] = None,
                new_snapshot_name: Optional[str] = None,
                new_snapshot_description: Optional[str] = None,
                **kwargs
            ) -> None:
        """
        Method to call a task to edit a snapshot.

        :param vm_id: Virtual datacenter identifier.
        :param pc_id: Remote pc identifier.
        :param snapshot_id: Snapshot identifier.
        :param new_snapshot_name: New name of snapshot to take.
        :param new_snapshot_description: New description of snapshot.
        """
        method = "vm.snapshot.edit.task"
        return await self.call(
            method, vm_id=vm_id, pc_id=pc_id, snapshot_id=snapshot_id,
            new_snapshot_name=new_snapshot_name, new_snapshot_description=new_snapshot_description,
            **kwargs)

    @overload
    async def vm_snapshot_get_list(
        self,
        *,
        pc_id: int,
    ) -> Dict[str, Any]:
        ...

    @overload
    async def vm_snapshot_get_list(
        self,
        *,
        vm_id: int,
    ) -> Dict[str, Any]:
        ...

    async def vm_snapshot_get_list(
                self,
                *,
                vm_id: Optional[int] = None,
                pc_id: Optional[int] = None,
                **kwargs
            ) -> Dict[str, Any]:
        """
        Method to call a task to retrieve a list of snapshots.

        :param vm_id: Virtual datacenter identifier.
        :param pc_id: Remote pc identifier.

        :return: Returns a dict with the following properties:

        **Properties**

        current_snapshot
            The current snapshot. See *Snapshot properties* for more info.
        snapshot_list
            A list of all snapshots. See *Snapshot properties* for more info.

        **Snapshot properties**

        power_state
            The state of the snapshot. :py:class:`PowerState`
        name
            The snapshot name.
        description
            The snapshot description.
        parent_id
            The id of the snapshot parent.
        id
            The snapshot id.
        mo_ref
            The unique identifier of the managed object in vSphere.
        """
        method = "vm.snapshot.get.list.task"
        return await self.call(method, vm_id=vm_id, pc_id=pc_id, **kwargs)

    @overload
    async def vm_snapshot_get_tree(
        self,
        *,
        pc_id: int,
    ) -> Dict[str, Any]:
        ...

    @overload
    async def vm_snapshot_get_tree(
        self,
        *,
        vm_id: int,
    ) -> Dict[str, Any]:
        ...

    async def vm_snapshot_get_tree(
                self,
                *,
                vm_id: Optional[int] = None,
                pc_id: Optional[int] = None,
                **kwargs
            ) -> Dict[str, Any]:
        """
        Method to call a task to retrieve snapshot tree.

        :param vm_id: Virtual datacenter identifier.
        :param pc_id: Remote pc identifier.

        :return: Returns a dict with the following properties:

        **Properties**

        current_snapshot
            The current snapshot. See *Snapshot properties* for more info.
        snapshot_list
            A list of all snapshots. See *Snapshot properties* for more info.

        **Snapshot properties**

        power_state
            The state of the snapshot. :py:class:`PowerState`
        name
            The snapshot name.
        description
            The snapshot description.
        id
            The snapshot id.
        mo_ref
            The unique identifier of the managed object in vSphere.
        """
        method = "vm.snapshot.get.tree.task"
        return await self.call(method, vm_id=vm_id, pc_id=pc_id, **kwargs)

    async def vm_snapshot_remove(
                self,
                *,
                vm_id: Optional[int] = None,
                pc_id: Optional[int] = None,
                snapshot_name: Optional[str] = None,
                snapshot_id: Optional[int] = None,
                remove_children: Optional[bool] = None,
                remove_all_snapshots: Optional[bool] = None,
                **kwargs
            ) -> None:
        """
        Method to call a task to remove snapshot from virtual machine.

        :param vm_id: Virtual datacenter identifier.
        :param pc_id: Remote pc identifier.
        :param snapshot_name: Name of snapshot to take.
        :param snapshot_id: Snapshot identifier.
        :param remove_children: Remove children snapshots.
        :param remove_all_snapshots: Remove all snapshots of virtual machine.
        """
        method = "vm.snapshot.remove.task"
        return await self.call(
            method, vm_id=vm_id, pc_id=pc_id, snapshot_name=snapshot_name, snapshot_id=snapshot_id,
            remove_children=remove_children, remove_all_snapshots=remove_all_snapshots, **kwargs)

    async def vm_snapshot_revert(
                self,
                *,
                vm_id: Optional[int] = None,
                pc_id: Optional[int] = None,
                snapshot_name: Optional[str] = None,
                snapshot_id: Optional[int] = None,
                **kwargs
            ) -> None:
        """
        Method to call a task to revert to snapshot from virtual machine.

        :param vm_id: Virtual datacenter identifier.
        :param pc_id: Remote pc identifier.
        :param snapshot_name: Name of snapshot to take.
        :param snapshot_id: Snapshot identifier.
        """
        method = "vm.snapshot.revert.task"
        return await self.call(
            method, vm_id=vm_id, pc_id=pc_id, snapshot_name=snapshot_name, snapshot_id=snapshot_id, **kwargs)

    async def vm_clone_task(
                self,
                *,
                parent_vm_id: int,
                parent_snapname: str,
                clone_role: enums.VirtualMachineRole,
                clone_type: enums.CloneType,
                clone_name: str,
                clone_datastore: Optional[str] = None,
                clone_storage_alloc: enums.CloneStorageAllocation = enums.CloneStorageAllocation.ONDEMAND,
                clone_vh_id: Optional[int] = None,
                clone_vhg_id: Optional[int] = None,
                clone_comments: Optional[str] = None,
                copy_bios_uuid: bool = False,
                **kwargs
            ) -> int:
        """
        Method to call a task to clone a virtual machine.

        :param parent_vm_id: Parent virtual machine identifier.
        :param parent_snapname: Parent snapshot name.
        :param clone_role: Cloned virtual machine role.
        :param clone_type: Cloned virtual machine type.
        :param clone_name: Cloned virtual machine name.
        :param clone_datastore: Datastore cloned virtual machine will reside on.
        :param clone_storage_alloc: Cloned virtual machine storage allocation.
        :param clone_vh_id: The virtual host the cloned virtual machine will reside on.
        :param clone_vhg_id: The virtual host group identifier. NOT SUPPORTED!
        :param clone_comments: Comments for the cloned virtual machine.

        :returns: Created task id
        """
        method = "vm.clone.task"
        args = {
            'clone_name': clone_name,
            'clone_type': clone_type,
            'clone_role': clone_role,
            'parent_vm_id': parent_vm_id,
            'copy_bios_uuid': copy_bios_uuid,
            'parent_snapname': parent_snapname,
        }
        args.update(kwargs)

        # if parent_snapname is not None:
        #     args['parent_snapname'] = parent_snapname
        if clone_datastore is not None:
            args['clone_datastore'] = clone_datastore
        if clone_vh_id is not None:
            args['clone_vh_id'] = clone_vh_id
        if clone_vhg_id is not None:
            args['clone_vhg_id'] = clone_vhg_id
        if clone_comments is not None:
            args['clone_comments'] = clone_comments

        return await self.call(method, **args)

    async def vm_inventory_remove_disk_task(
                self,
                *,
                vm_id: int,
                **kwargs
            ) -> None:
        """
        Remove virtual machine from NETLAB+ inventory, datacenter, and disk.

        **WARNING: this method permanantly deletes the virtual machine files
        from disk.  This operation cannot be undone.**

        :param vm_id: Virtual machine identifier.
        """
        method = 'vm.inventory.remove.disk.task'

        args = {
            'vm_id': vm_id,
        }

        args.update(kwargs)

        return await self.call(method, **args)

    async def vm_inventory_remove_datacenter_task(
                self,
                *,
                vm_id: int,
                **kwargs
            ) -> None:
        """
        Remove virtual machine from NETLAB+ inventory and datacenter.
        Does NOT remove VM from disk.

        :param int vm_id: Virtual machine identifier.
        """
        method = "vm.inventory.remove.datacenter.task"

        args = {
            'vm_id': vm_id,
        }
        args.update(kwargs)

        return await self.call(method, **args)

    async def vm_inventory_remove_local(
                self,
                *,
                vm_id,
                **kwargs
            ) -> None:
        """
        Remove virtual machine from NETLAB+ inventory only.
        Does NOT remove VM from datacenter or disk.

        :param vm_id: Virtual machine identifier.
        """
        method = "vm.inventory.remove.local"

        args = {
            'vm_id': vm_id,
        }
        args.update(kwargs)

        return await self.call(method, **args)

    @minimum_version('18.2.0')
    async def vm_license_list(
                self,
                *,
                vdc_id: Optional[int] = None,
                vh_id: Optional[int] = None,
                vl_server: Optional[int] = None,
                vl_key: Optional[int] = None,
                vl_type: Optional[enums.VirtualMachineLicenseType] = None,
                **kwargs
            ) -> List[Dict[str, Any]]:
        """
        This method returns a list of cached VMware vCenter and server licenses.

        This license cache is populated by any of the following events:
        - every 12 hours per scheduled task
        - when datacenters are added or removed
        - when virtual machine host servers are added or removed
        - by calling API method :py:meth:`VmApiMixin.vm_license_update_task`

        The cache updates each time licenses are scanned.  If a vCenter server for a datacenter is
        unreachable during the scan, the licenses managed by that vCenter will be missing from the
        cache until the vCenter can be reached on a subsequent scan.

        :param vdc_id: return licenses for datacenter identified by **vdc_id** in NETLAB+
        :param vh_id: return license for vm host identified by **vh_id** in NETLAB+
        :param vl_server: return license with hostname or IP address matching **vl_server** in vCenter
        :param vl_key: return license with key matching **vl_key** in vCenter
        :param vl_type: License type.

        :returns: A list of dictionaries with the following properties.

        **Properties**

        vdc_id
            NETLAB+ datacenter identifier associated with license.
        vh_id
            If not `None`, the NETLAB+ vm host identifier associated with license.
        vl_type
            License type. :py:class:`netlab.enums.VirtualMachineLicenseType`
        vl_server
            Hostname or IP address of the server as registered in vCenter.
        vl_product
            Licensed product name.
        vl_key
            Licnese key.
        vl_units
            Number of units licensed (i.e. processors) depending on server type.
        vl_expiration
            The expiration date and time of the license (UTC) or `None` if no expiration.
        vl_last_udpate
            The date and time this license was last queried and cached.
        """

        method = "vm.license.list"

        params: Dict[str, Any] = {}
        params.update(**kwargs)

        if vdc_id is not None:
            params['vdc_id'] = vdc_id
        if vh_id is not None:
            params['vh_id'] = vh_id
        if vl_server is not None:
            params['vl_server'] = vl_server
        if vl_key is not None:
            params['vl_key'] = vl_key
        if vl_type is not None:
            params['vl_type'] = vl_type

        return await self.call(method, **params)

    @minimum_version('18.2.0')
    async def vm_license_update_task(self, **kwargs) -> Dict[str, Any]:
        """
        This method scans each vCenter server in all NETLAB+ datacenters and returns the VMware
        license status of VMware vCenter and VMware host servers.

        * The license information is cached for quick retreival using subsequent calls to
          :py:meth:`VmApiMixin.vm_license_list`.

        * It is usually not necessary to call vm.liense.update.task directly.  It is automatically run:
            - every 12 hours per scheduled task
            - when datacenters are added or removed
            - when virtual machine host servers are added or removed

        * The license cache is cleared updated each time vm.license.udpate.task is run.
            - If a vCenter server for a datacenter is unreachable during the scan, the licenses
              managed by that vCenter will be missing from the cache until the vCenter can be
              reached on a subsequent scan.

        * This task ignores licenses that are not associated with NETLAB+ servers.

        :returns: A dictionary with the following properties.

        **Properties**

        servers_tried
            The number of vCenter servers that contact was attempted.
        servers_reporting
            The number of vCenter servers that responded with license information.
        license_count
            The number of licenses that were found and cached.
        licenses
            Information about each licence (see below).
        earliest_expiration
            The earliest license expiration at time of call. Can be `None`.

        **License Properties**

        vdc_id
            NETLAB+ datacenter identifier associated with license.
        vh_id
            If not `None`, the NETLAB+ vm host identifier associated with license.
        vl_type
            License type. :py:class:`netlab.enums.VirtualMachineLicenseType`
        vl_server
            Hostname or IP address of the server as registered in vCenter.
        vl_product
            Licensed product name.
        vl_key
            Licnese key.
        vl_units
            Number of units licensed (i.e. processors) depending on server type.
        vl_expiration
            The expiration date and time of the license (UTC) or `None` if no expiration.
        vl_last_udpate
            The date and time this license was last queried and cached.
        """

        method = "vm.license.update.task"
        return await self.call(method, **kwargs)

    async def vm_host_perf_realtime_list(
                self,
                *,
                vh_id: Optional[List[int]] = None,
                **kwargs,
            ) -> List[Dict[str, Any]]:
        """
        This method returns real-time virtual machine host performance data for one or more hosts.

        :params vh_id: NETLAB+ host identifier of hosts to return. If omitted or empty, all
            hosts are returned.
        :returns: A dictionary with the following properties.

        **Recommended Usage**

        1. Obtain a list of hosts systems from vm.host.list.
        2. Call this method to obtain real-time performance data for one or more hosts. You
           may optionally specify which hosts to return by passing an array or CSV list of
           host IDs; long running applications that specify vh_id should always be prepared
           to handle this exception in case a host host is deleted.
        3. Real-time data is available every 20 seconds with observations for T+00, T+20, and
           T+40 seconds of every minute. For optimal results, call this method at T+05, T+25,
           and T+45 seconds; calling system should have accurate system clock (i.e. NTP). To
           avoid unnecessary overhead on the NETLAB+ server API, calls to this method should
           be at least 20 seconds apart.

        **Properties**

        vh_id
            The NETLAB+ host identifier of the associated host.
        data_available
            real-time performance properties are available for this host.
        vhp_time
            The ending date and time (UTC) of the performance interval (UTC).
        age_sec
            The age of the data in second from time of call. If greater than 60, data will
            be considered too stale to be real-time, and data_available will be 0. If None,
            no data was available for host, data_available will be 0 and vhp_time will be None.

        """
        method = "vm.host.perf.realtime.list"

        params = {}

        if vh_id is not None:
            params['vh_id'] = ','.join(map(str, vh_id))

        return await self.call(method, **params, **kwargs)



from datetime import date
from typing import Any, Dict, List, Literal, Optional, Union, cast

from .. import enums
from ._client_protocol import ClientProtocol


class ClassApiMixin(ClientProtocol):
    "Methods for working with classes."

    class_default_props = ['cls_id', 'cls_name', 'com_id']
    class_props = ['cls_change_ex', 'cls_def_pw_account', 'cls_def_pw_console', 'cls_def_pw_enable',
                   'cls_email_logs', 'cls_end_date', 'cls_id', 'cls_lab_limit', 'cls_max_slots_per_res',
                   'cls_min_hours_btw_res', 'cls_name', 'cls_no_delete', 'cls_retain_ilt', 'cls_retain_period',
                   'cls_retain_st', 'cls_self_sched', 'cls_start_date', 'cls_team_sched', 'cls_uuid', 'com_id',
                   'enrollment', 'leads']

    roster_default_props = ['acc_id', 'acc_user_id', 'acc_email', 'lead']
    roster_props = ['acc_display_name', 'acc_email', 'acc_full_name', 'acc_id', 'acc_last_login', 'acc_logins',
                    'acc_sort_name', 'acc_time_last_access', 'acc_type', 'acc_user_id', 'cls_id', 'lead', 'ros_team']

    async def class_list(
                self,
                *,
                com_id: Optional[int] = None,
                member: bool = False,
                properties: Union[List[str], Literal['default', 'all']] = "default",
                **kwargs
            ) -> List[Dict[str, Any]]:
        """
        This method allows you to retrieve a list of classes and their properties.

        :param com_id: Community ID.
        :param member: If true, get list of classes the caller is in.
        :param properties: List of properties to retrieve. See Properties section for list of available properties.

        :return: On success, returns an object with the requested properties:

        **Properties**

        cls_change_ex
            Allow student to change exercise.
        cls_def_pw_account
            Class default account password.
        cls_def_pw_console
            Class default console password.
        cls_def_pw_enable
            Class default enable password.
        cls_email_logs
            Email reservation logs for real equipment pods. :py:class:`netlab.enums.ClassEmailLogs`
        cls_end_date
            Class end date.
        cls_id
            Unique class identifier.
        cls_lab_limit
            Lab limit settings. :py:class:`netlab.enums.ClassLabLimit`
        cls_max_slots_per_res
            Maximum number of slots per reservation.
        cls_min_hours_btw_res
            Minimum number of hours between reservations.
        cls_name
            Unique class name per community.
        cls_no_delete
            Prevent from deletion.
        cls_retain_ilt
            Retain instructor led logs.
        cls_retain_period
            Number of days to retain class logs.
        cls_retain_st
            Retain student logs.
        cls_self_sched
            Allow students to self schedule.
        cls_start_date
            Class start date.
        cls_team_sched
            Allow teams to schedule.
        cls_uuid
            Class UUID.
        com_id
            Unique community identifier.
        enrollment
            Count of those enrolled.
        leads
            List of leads and properties sorted by acc_sort_name.
        """
        method = "class.list"
        if properties == 'all':
            properties = self.class_props
        if properties == 'default':
            properties = self.class_default_props
        return await self.call(method, com_id=com_id, member=member, properties=properties, **kwargs)

    async def class_add(
                self,
                *,
                cls_name: str,
                com_id: int,
                cls_change_ex: bool = True,
                cls_def_pw_account: Optional[str] = None,
                cls_def_pw_console: Optional[str] = None,
                cls_def_pw_enable: Optional[str] = None,
                cls_email_logs: enums.ClassEmailLogs = enums.ClassEmailLogs.NO,
                cls_end_date: Optional[date] = None,
                cls_lab_limit: enums.ClassLabLimit = enums.ClassLabLimit.ENFORCE,
                cls_max_slots_per_res: Optional[int] = None,
                cls_min_hours_btw_res: Optional[int] = None,
                cls_no_delete: bool = False,
                cls_retain_ilt: bool = False,
                cls_retain_period: Optional[date] = None,
                cls_retain_st: bool = True,
                cls_self_sched: bool = True,
                cls_start_date: Optional[date] = None,
                cls_team_sched: bool = False,
                **kwargs
            ) -> int:
        """
        This method allows you to add a class to the NETLAB+ system.

        :param cls_name: Class name.
        :param com_id: Community ID.

        :param cls_change_ex: Allow student to change exercise.
        :param cls_def_pw_account: Class default account password.
        :param cls_def_pw_console: Class default console password.
        :param cls_def_pw_enable: Class default enable password.
        :param cls_email_logs: Email reservation logs for real equipment pods.
        :param cls_end_date: Class end date.
        :param cls_lab_limit: Class lab time limit.
        :param cls_max_slots_per_res: Maximum number of slots per reservation.
        :param cls_min_hours_btw_res: Minimum hours between reservations.
        :param cls_no_delete: Prevent from deletion.
        :param cls_retain_ilt: Retain instructor led logs.
        :param cls_retain_period: The number of days to retain logs for.
        :param cls_retain_st: Retain student logs.
        :param cls_self_sched: Allow students to self schedule.
        :param cls_start_date: Class start date.
        :param cls_team_sched: Allow teams to schedule.

        :return: On success, returns the **cls_id** of the newly created class.
        """

        method = "class.add"
        params = {'cls_name': cls_name, 'com_id': com_id, 'cls_change_ex': cls_change_ex,
                  'cls_email_logs': cls_email_logs, 'cls_lab_limit': cls_lab_limit, 'cls_no_delete': cls_no_delete,
                  'cls_retain_ilt': cls_retain_ilt, 'cls_retain_st': cls_retain_st, 'cls_self_sched': cls_self_sched,
                  'cls_team_sched': cls_team_sched}

        if cls_def_pw_account:
            params['cls_def_pw_account'] = cls_def_pw_account
        if cls_def_pw_console:
            params['cls_def_pw_console'] = cls_def_pw_console
        if cls_def_pw_enable:
            params['cls_def_pw_enable'] = cls_def_pw_enable

        if cls_start_date:
            params['cls_start_date'] = cls_start_date
        if cls_end_date:
            params['cls_end_date'] = cls_end_date

        if cls_max_slots_per_res:
            params['cls_max_slots_per_res'] = cls_max_slots_per_res
        if cls_min_hours_btw_res:
            params['cls_min_hours_btw_res'] = cls_min_hours_btw_res

        if cls_retain_period:
            params['cls_retain_period'] = cls_retain_period

        params.update(kwargs)
        return int(await self.call(method, **params))

    async def class_get(
                self,
                *,
                cls_id: int,
                properties: Union[List[str], Literal['all', 'default']] = "default",
                **kwargs
            ) -> Dict[str, Any]:
        """
        This method allows you to retrieve a single class and properties.

        :param int cls_id: Class ID.
        :param properties: List of properties to retrieve. Defaults to "all".
        :return: On success, returns an object with the requested properties:

        **Properties**

        cls_change_ex
            Allow student to change exercise.
        cls_def_pw_account
            Class default account password.
        cls_def_pw_console
            Class default console password.
        cls_def_pw_enable
            Class default enable password.
        cls_email_logs
            Email reservation logs for real equipment pods. :py:class:`netlab.enums.ClassEmailLogs`
        cls_end_date
            Class end date.
        cls_id
            Unique class identifier.
        cls_lab_limit
            Lab limit settings. :py:class:`netlab.enums.ClassLabLimit`
        cls_max_slots_per_res
            Maximum number of slots per reservation.
        cls_min_hours_btw_res
            Minimum number of hours between reservations.
        cls_name
            Unique class name per community.
        cls_no_delete
            Prevent from deletion.
        cls_retain_ilt
            Retain instructor led logs.
        cls_retain_period
            Number of days to retain class logs.
        cls_retain_st
            Retain student logs.
        cls_self_sched
            Allow students to self schedule.
        cls_start_date
            Class start date.
        cls_team_sched
            Allow teams to schedule.
        cls_uuid
            Class UUID.
        com_id
            Unique community identifier.
        enrollment
            Count of those enrolled.
        leads
            List of leads and properties sorted by acc_sort_name.
        """

        method = "class.get"
        if properties == 'all':
            properties = self.class_props
        if properties == 'default':
            properties = self.class_default_props
        return await self.call(method, cls_id=cls_id, properties=properties, **kwargs)

    async def class_update(
                self,
                *,
                cls_id: int,
                cls_change_ex: Optional[bool] = None,
                cls_def_pw_account: Optional[str] = None,
                cls_def_pw_console: Optional[str] = None,
                cls_def_pw_enable: Optional[str] = None,
                cls_email_logs: Optional[enums.ClassEmailLogs] = None,
                cls_end_date: Optional[date] = None,
                cls_lab_limit: Optional[enums.ClassLabLimit] = None,
                cls_max_slots_per_res: Optional[int] = None,
                cls_min_hours_btw_res: Optional[int] = None,
                cls_name: Optional[str] = None,
                cls_no_delete: Optional[bool] = None,
                cls_retain_ilt: Optional[bool] = None,
                cls_retain_period: Optional[int] = None,
                cls_retain_st: Optional[bool] = None,
                cls_self_sched: Optional[bool] = None,
                cls_start_date: Optional[date] = None,
                cls_team_sched: Optional[bool] = None,
                cls_ext_slots_per_res: Optional[Union[int, enums.ClassExtensionSlots]] = None,
                **kwargs
            ) -> Literal["OK"]:
        """
        This method allows you to update a class and its properties.

        :param cls_id: Unique class identifier.
        :param cls_change_ex: Allow student to change exercise.
        :param cls_def_pw_account: Class default account password.
        :param cls_def_pw_console: Class default console password.
        :param cls_def_pw_enable: Class default enable password.
        :param cls_email_logs: Email reservation logs for real equipment pods.
        :param cls_end_date: Class end date.
        :param cls_lab_limit: Class lab time limit.
        :param cls_max_slots_per_res: Maximum number of slots per reservation.
        :param cls_min_hours_btw_res: Minimum hours between reservations.
        :param cls_name: Class name.
        :param cls_no_delete: Prevent from deletion.
        :param cls_retain_ilt: Retain instructor led logs.
        :param cls_retain_period: The number of days to retain logs for.
        :param cls_retain_st: Retain student logs.
        :param cls_self_sched: Allow students to self schedule.
        :param cls_start_date: Class start date.
        :param cls_team_sched: Allow teams to schedule.
        :param cls_ext_slots_per_res: Allow extensions to reservations for students and teams.
            Instructors are not restricted to this limit.
        """

        method = "class.update"
        params: Dict[str, Any] = {'cls_id': cls_id}

        if cls_name:
            params['cls_name'] = cls_name
        if cls_start_date:
            params['cls_start_date'] = cls_start_date
        if cls_end_date:
            params['cls_end_date'] = cls_end_date
        if cls_def_pw_account:
            params['cls_def_pw_account'] = cls_def_pw_account
        if cls_def_pw_console:
            params['cls_def_pw_console'] = cls_def_pw_console
        if cls_def_pw_enable:
            params['cls_def_pw_enable'] = cls_def_pw_enable
        if cls_email_logs:
            params['cls_email_logs'] = cls_email_logs
        if cls_lab_limit:
            params['cls_lab_limit'] = cls_lab_limit
        if cls_no_delete is not None:
            params['cls_no_delete'] = cls_no_delete
        if cls_retain_ilt is not None:
            params['cls_retain_ilt'] = cls_retain_ilt
        if cls_retain_st is not None:
            params['cls_retain_st'] = cls_retain_st
        if cls_retain_period:
            params['cls_retain_period'] = cls_retain_period
        if cls_self_sched is not None:
            params['cls_self_sched'] = cls_self_sched
        if cls_team_sched is not None:
            params['cls_team_sched'] = cls_team_sched
        if cls_change_ex is not None:
            params['cls_change_ex'] = cls_change_ex
        if cls_min_hours_btw_res:
            params['cls_min_hours_btw_res'] = cls_min_hours_btw_res
        if cls_max_slots_per_res:
            params['cls_max_slots_per_res'] = cls_max_slots_per_res
        if cls_ext_slots_per_res is not None:
            params['cls_ext_slots_per_res'] = cls_ext_slots_per_res

        params.update(kwargs)
        return await self.call(method, **params)

    async def class_remove(
                self,
                *,
                cls_id: int,
                delete_students: bool = False,
                **kwargs
            ) -> None:
        """
        This method removes a class from the NETLAB+ system.

        :param cls_id: Local system class.
        :param delete_students: `False` Do not delete student accounts. `True` Delete student accounts
            that are exclusively in this class. Student accounts that are members of multiple classes are not deleted.
        """

        method = "class.remove"
        return await self.call(method, cls_id=cls_id, delete_students=delete_students, **kwargs)

    async def class_roster_add(
                self,
                *,
                acc_id: str,
                cls_id: str,
                lead: bool = False,
                ros_team: Optional[str] = None,
                **kwargs
            ) -> Literal["OK"]:
        """
        This method allows you to add a member to a class roster.

        :param acc_id: Unique account identifier.
        :param cls_id: Unique class identifier.
        :param lead: Add as lead.
        :param ros_team: Assign user to a team. Single character, A-Z.
        """

        method = "class.roster.add"
        params = {'acc_id': acc_id, 'cls_id': cls_id, 'lead': lead}

        if ros_team:
            params['ros_team'] = ros_team

        params.update(kwargs)
        return await self.call(method, **params)

    async def class_roster_list(
                self,
                *,
                cls_id: int,
                leads: bool = False,
                **kwargs
            ):
        """
        This method lists users in the class roster.

        :param cls_id: Local system class.
        :param leads: `None` Return all users in the roster (leads and learners).
            `True` Return lead instructors only. `False` Return learners only.

        :return: On success, returns an object with the requested properties:

        **Properties**

        acc_display_name
            Account name people will see this name in discussions, messages and comments.
        acc_email
            Account email.
        acc_id
            Local system account identifier.
        acc_full_name
            Account full name.
        acc_last_login
            Timestamp of last account login. Possibly `None`.
        acc_logins
            Number of times a user has logged in.
        acc_sort_name
            Account name as it appears in sorted lists.
        acc_time_last_access
            Timestamp of last access. Possibly `None`.
        acc_type
            Account type. :py:class:`netlab.enums.AccountType`
        acc_user_id
            Account user identifier.
        cls_id
            Unique class identifier.
        lead
            If user is a lead instructor.
        ros_team
            Assignment of user to a team (designated by team letter). `None` if user is not assigned to a team.
        """

        method = "class.roster.list"
        return await self.call(method, cls_id=cls_id, leads=leads, **kwargs)

    async def class_roster_remove(
                self,
                *,
                acc_id: int,
                cls_id: str,
                **kwargs
            ) -> None:
        """
        This method removes a user from the class roster.

        :param acc_id: Account identifier of user to be removed.
        :param cls_id: Unique class identifier.
        """

        method = "class.roster.remove"
        return await self.call(method, acc_id=acc_id, cls_id=cls_id, **kwargs)

    async def class_roster_get(
                self,
                *,
                cls_id: int,
                acc_id: Optional[int],
                **kwargs,
            ) -> Dict[str, Any]:
        """
        This method returns class roster information for a single user.

        :param cls_id: Local system class identifier.
        :param acc_id: Returns roster information for the user specified by acount ID acc_id.
            If acc_id is not provided, roster information related to the requester is returned.
        :return: On success, returns an object with the requested properties:

        **Properties**

        acc_display_name
            Account name people will see this name in discussions, messages and comments.
        acc_email
            Account email.
        acc_id
            Local system account identifier.
        acc_full_name
            Account full name.
        acc_last_login
            Timestamp of last account login.
        acc_logins
            Number of times a user has logged in.
        acc_sort_name
            Account name as it appears in sorted lists.
        acc_time_last_access
            Timestamp of last access.
        acc_type
            Account type.
        acc_user_id
            Account user identifier.
        cls_id
            Unique class identifier.
        lead
            User is a lead instructor.
        ros_team
            Team identifier [A-Z] or `None` if user is not assigned to a team.
        """
        method = 'class.roster.get'

        return await self.call(method, acc_id=acc_id, cls_id=cls_id, **kwargs)

    async def class_roster_team_update(
                self,
                *,
                cls_id: int,
                ros_team: Optional[str] = None,
                roster_acc_id: List[int],
                **kwargs,
            ) -> str:
        """
        This method will create and update teams (groups of students) for a given class roster.

        :param cls_id: Local system class identifier.
        :param ros_team: A single uppercase letter [A,Z] as the designation of the team to add
            the student(s) to. If this value is `None`, the student(s) are assigned to no team
            (unassigned).
        :param roster_acc_id: A list of account IDs to assign to the specified team.
        """

        method = 'class.roster.team.update'

        acc_id_csv = ','.join(map(str, roster_acc_id))

        return await self.call(method, cls_id=cls_id, ros_team=ros_team, roster_acc_id=acc_id_csv, **kwargs)

    async def class_content_add(
        self,
        *,
        cls_id: int,
        con_id: str,
        **kwargs,
    ) -> None:
        """
        This method allows you to add content to a class.

        :param cls_id: Class identifier.
        :param con_id: Content identifier.
        """

        method = "class.content.add"
        return await self.call(
            method,
            cls_id=cls_id,
            con_id=con_id,
            **kwargs,
        )

    async def class_content_availability(
        self,
        *,
        cls_id: int,
        **kwargs,
    ) -> List[Dict[str, Any]]:
        """
        This method provides a list of available class content.

        :param cls_id: Unnique class identifier.

        :return: On success, returns a list of objects with the requested properties:

        **Properties**

        con_access
            Content is Globally accessible or Private.
        con_aid
            Content author ID.
        con_author
            Content author name.
        con_build
            Build number.
        con_copyright
            Content copyright.
        con_desc
            Content description.
        con_gid
            Unique content identifier.
        con_global
            Is content globally accessible or not.
        con_id
            Unique content identifier.
        con_managed
            Is the content managed or not.
        con_name
            Name of content.
        con_notes
            Content notes.
        con_org
            Content organization.
        con_origin
            Source of the content.
        con_pt_ids
            Comma separated list of required Pod Types.
        con_url
            Content URL.
        """
        method = "class.content.availability"

        return await self.call(
            method,
            cls_id=cls_id,
            **kwargs,
        )

    async def class_content_list(
        self,
        *,
        cls_id: int,
        properties: Union[Literal['all'], List[str]] = 'all',
        **kwargs,
    ) -> List[Dict[str, Any]]:
        """
        This method provides a list of content associated with a class.

        :param cls_id: Unique class identifier.
        :param properties: Array of properties to retrieve.

        :return: On success, returns a list of objects with the requested properties:

        **Properties**

        con_access
            Content is Globally accessible or Private.
        con_aid
            Content author ID.
        con_author
            Content author name.
        con_build
            Build number.
        con_copyright
            Content copyright.
        con_desc
            Content description.
        con_gid
            Unique content identifier.
        con_global
            Is content globally accessible or not.
        con_id
            Unique content identifier.
        con_managed
            Is the content managed or not.
        con_name
            Name of content.
        con_notes
            Content notes.
        con_org
            Content organization.
        con_origin
            Source of the content.
        con_pt_ids
            Comma separated list of required Pod Types.
        con_url
            Content URL.
        """
        method = "class.content.list"

        if properties == 'all':
            properties = cast(Literal['all'], None)

        return await self.call(
            method,
            cls_id=cls_id,
            properties=properties,
            **kwargs,
        )

    async def class_content_remove(
        self,
        *,
        cls_id: int,
        con_id: str,
        **kwargs,
    ) -> None:
        """
        This method removes content from a class.

        :param cls_id: Class identifier.
        :param con_id: Content identifier.
        """
        method = "class.content.remove"

        return await self.call(
            method,
            cls_id=cls_id,
            con_id=con_id,
            **kwargs,
        )



import asyncio

from netlab import enums
from netlab.async_client import NetlabClient

POD_RANGE = range(900, 910)


async def remove_pod(client, pod_number):
    await client.pod_state_change(pod_id=pod_number, state=enums.PodState.OFFLINE)
    await client.pod_remove_task(pod_id=pod_number, remove_vms=enums.RemoveVMS.DISK)

    return pod_number


async def main():
    async with NetlabClient() as client:
        for pod_task in asyncio.as_completed([
            remove_pod(client, i) for i in POD_RANGE
        ]):
            pod_name = await pod_task

            print(f'Removed pod: {pod_name}')


if __name__ == '__main__':
    asyncio.run(main())



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['netlab',
 'netlab.api',
 'netlab.cli',
 'netlab.datatypes',
 'netlab.enums',
 'netlab.errors']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.1.7,<9.0.0',
 'packaging>=21.3',
 'rich>=13.8.1,<14.0.0',
 'tabulate>=0.8,<0.9',
 'typing-extensions>=4.12.2,<5.0.0']

entry_points = \
{'console_scripts': ['netlab = netlab.cli:main']}

setup_kwargs = {
    'name': 'netlab',
    'version': '23.1.0',
    'description': 'NETLAB+ VE SDK for Python',
    'long_description': '# OVERVIEW\n\n[![pipeline status](https://gitlab.netdevgroup.com/dev/netlab-ve-api-py/badges/master/pipeline.svg)](https://gitlab.netdevgroup.com/dev/netlab-ve-api-py/-/pipelines?scope=all&ref=master)\n[![coverage report](https://gitlab.netdevgroup.com/dev/netlab-ve-api-py/badges/master/coverage.svg)](https://gitlab.netdevgroup.com/dev/netlab-ve-api-py/-/commits/master)\n[![downloads](https://img.shields.io/endpoint?url=https://ndg.tech/api/badges/netlab-py-latest/)](https://portal.netdevgroup.com/staff/tech/url/netlab-py-latest)\n[![python versions](https://shields.io/badge/python-3.7%20|%203.8%20|%203.9%20|%203.10-brightgreen)](https://gitlab.netdevgroup.com/dev/netlab-ve-api-py/-/blob/master/tox.ini)\n\nThe goal of this project is to provide a mechanism for NETLAB+ administrators to\nprogramatically manage his or her NETLAB+ appliance.\n\nPython is the language of choice for creating API calls directly to NETLAB+.\nWith this package you will be able to manipulate (add/create, update, delete,\nget/list) user\'s, pod\'s, classes, vm\'s, reservations and more.\n\n\n# INSTALL\n\n    pip install https://ndg.tech/netlab-py-latest\n\n\n# LICENSE\n\nMIT License\n\nCopyright (c) 2017 Network Development Group, Inc\n\nPermission is hereby granted, free of charge, to any person obtaining a copy of this software and associated\ndocumentation files (the "Software"), to deal in the Software without restriction, including without limitation\nthe rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and\nto permit persons to whom the Software is furnished to do so, subject to the following conditions:\n\nThe above copyright notice and this permission notice shall be included in all copies or substantial portions of\nthe Software.\n\nTHE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE\nWARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS\nOR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR\nOTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.\n',
    'author': 'Aaron Schif',
    'author_email': 'aschif@netdevgroup.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://www.netdevgroup.com/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9.0,<4',
}


setup(**setup_kwargs)



import asyncio
import json
import sys
from collections import defaultdict
from typing import DefaultDict

from paepy.ChannelDefinition import CustomSensorResult  # type: ignore

from netlab.async_client import NetlabClient


async def main():
    # interpret first command line parameter as json object
    if len(sys.argv) == 1:
        # for command line debugging, use the default system
        sys_args = '{"params": "default"}'
    else:
        sys_args = sys.argv[1]
    prtg = json.loads(sys_args)
    system = prtg.get('params', "default").lower()

    # PRTG runs sensors as a different user. Specify the Administrator user's netlab config file here
    async with NetlabClient(system, config_path='C:/Users/Administrator/.netlab/config.json') as client:

        # query data from netlab api
        total: DefaultDict[int, int] = defaultdict(int)
        used: DefaultDict[int, int] = defaultdict(int)

        for pod in await client.pod_list():
            total[pod['pt_id']] += 1
            if pod['pod_res_id']:
                used[pod['pt_id']] += 1

        pod_types = {pt['pt_id']: pt['pt_name'] for pt in await client.pod_types_list()}

        # create sensor result
        result = CustomSensorResult("OK")

        # add primary channel for total pods reserved
        total_pods = sum(total.values())
        used_pods = sum(used.values())
        total_used_pct = (used_pods / total_pods) * 100
        result.add_channel(channel_name="Overall Percent Used", primary_channel=True, unit='Percent',
                           value=int(total_used_pct), is_float=False)

        # additional channel for each individual pod type percent, used and total
        for pt_id, total_count in total.items():
            used_count = used[pt_id]
            used_pct = (used_count / total_count) * 100
            result.add_channel(channel_name="% - {}".format(pod_types[pt_id]), unit="Percent",
                               value=int(used_pct), is_float=False)
            result.add_channel(channel_name="Total - {}".format(pod_types[pt_id]), unit="Count",
                               value=total_count, is_float=False)
            result.add_channel(channel_name="Used - {}".format(pod_types[pt_id]), unit="Count",
                               value=used_count, is_float=False)

        # print sensor result to stdout
        print(result.get_json_result())

asyncio.run(main())



from typing import Any, AsyncGenerator, Dict, List, Literal, Optional, Union

from .. import enums
from ..utils import minimum_version
from ._client_protocol import ClientProtocol


class UserApiMixin(ClientProtocol):
    user_account_default_props = ['acc_id', 'acc_user_id', 'acc_email', 'acc_type']
    user_account_list_props = ['acc_can_login', 'acc_display_name', 'acc_email', 'acc_full_name', 'acc_id',
                               'acc_last_ip', 'acc_last_login', 'acc_last_ua',
                               'acc_logins', 'acc_privs', 'acc_pw_change', 'acc_session_time_started', 'acc_sort_name',
                               'acc_sys', 'acc_time_created', 'acc_time_last_access', 'acc_type', 'acc_user_id',
                               'acc_uuid', 'com_id', 'tz_id']
    user_account_get_props = user_account_list_props + ['com_full_name', 'tz_iana', 'tz_name', 'acc_nlx_terms',
                                                        'acc_npd_terms', 'acc_pw_console', 'acc_pw_enable',
                                                        'accman_list_com_id', 'date_format', 'time_format',
                                                        'first_weekday', 'page_length']

    user_account_privs = ['COMWIDE', 'SYSWIDE', 'POD_DESIGNER', 'LAB_DESIGNER']

    community_default_props = ['com_id', 'com_full_name']
    community_props = ['com_alt_banner', 'com_alt_banner_height', 'com_alt_banner_width', 'com_enabled',
                       'com_full_name', 'com_id', 'com_max_slots_per_res', 'com_min_hours_btw_res',
                       'com_mynetlab_news', 'com_mynetlab_welcome']

    async def user_account_list(
                self,
                *,
                com_id: Optional[int] = None,
                acc_type: Optional[enums.AccountType] = None,
                properties: Union[List[str], Literal['all', 'default']] = "default",
                **kwargs
            ) -> List[Dict[str, Any]]:
        """
        Query a list of the user accounts in the NETLAB system. The list method
        is only restricted by community id and account types. The properties
        only identify those properties or fields that will be returned via
        the underlying query.

        :param com_id: Community ID
        :param acc_type: User account type.
        :param properties: Properties consist of any combination of properties listed below under Properties.
                           You may also use **"all"** to include all properties, or **"default"** for a smaller subset
                           of properties.

        :return: A list of dictionaries (a set of records) where the dictionaries contain requested properties
            per user.

        **Properties**

        acc_can_login
            User allowed to login.
        acc_display_name
            Account name people will see this name in discussions, messages and comments.
        acc_email
            Account email.
        acc_fs_id
            Filesystem identifier of the account root folder.
        acc_full_name
            Account full name.
        acc_id
            Account identifier.
        acc_last_ip
            Last IP address of user.
        acc_last_login
            Datetime of last account login.
        acc_last_ua
            User agent string of the browser when logging in from the UI.
        acc_logins
            Number of account logins.
        acc_privs
            Supplemental account privileges for instructors only. :py:class:`netlab.enums.AccountPrivileges`
        acc_pw_change
            Force user to change password.
        acc_session_time_started
            Datetime of session start time.
        acc_sort_name
            Account name as it appears in sorted lists.
        acc_sys
            Mark as system account.
        acc_time_created
            Datetime account was created.
        acc_time_last_access
            Datetime of last access.
        acc_type
            Account type. :py:class:`netlab.enums.AccountType`
        acc_user_id
            Unique system-wide login identifier.
        acc_uuid
            Globally unique account ID.
        com_id
            Unique community identifier.
        tz_id
            Time zone identifier.
        """

        if properties == 'all':
            properties = self.user_account_list_props
        if properties == 'default':
            properties = self.user_account_default_props

        filters: Dict[str, Any] = {}
        if com_id is not None:
            filters['com_id'] = com_id
        if acc_type is not None:
            filters['acc_type'] = acc_type

        result = []

        async for user in self.user_account_search_iter(properties=properties, filter=filters, limit=500):
            result.append(user)
        return result

    @minimum_version('17.1.4')
    async def user_account_search_iter(
                self,
                *,
                properties: List[str] = ['acc_id'],
                order: Optional[Union[str, List[str]]] = None,
                filter: Any = None,  # TODO
                limit: int = 100,
            ) -> AsyncGenerator[Dict[str, Any], None]:
        """
        This method queries and/or retrieves user accounts.

        Gets all users from server and then performs filtering locally. Any property from
        :meth:`~netlab.api.UserApiMixin.user_account_list` can be filtered upon. Multiple
        **filter** will be combined using AND. Returns zero or more user accounts that match filter query.

        :param properties: Properties consist of any combination of properties listed below under Properties.
        :param order: Sort order.
        :param filter: Search criteria. See below for how to use filters.
        :param limit: Page size collected in a single request from the server. This can be used to tune performace
            over large record sets. Maximum 500.

        :return: A series of user accounts that match **filter**. This method returns an iterator, which can be
            used in for loops like a list. See
            [the offical Python documenation](https://docs.python.org/3.6/library/stdtypes.html#iterator-types) for more
            infomation.

        **Examples**

        Collect the account id and account user id with **acc_id** that is 100031.

            Client().user_account_search_iter(properties=['acc_id', 'acc_user_id'], filter={'acc_id': 100031})

        Collect first page of accounts, ordered by **acc_id**.

            Client().user_account_search_iter(properties=['acc_id', 'acc_user_id'], order='acc_id')

        Return how many users can login.

            Client().user_account_search_iter(properties=['acc_id'], filter={'acc_can_login': True,
                                              'communities.com_id': 1})


        **Properties**

        The following properties may be used in this method. The text in bold indicates the contexts in which the
        property may be used (see contexts below). The text following the | character describes the property.

        Properties in **communities** and **timezones** must be fully qualified (i.e. `timezones.tz_name`). Properties
        starting with **accounts** my be specified without namespace (i.e. `acc_full_name` or `accounts.acc_full_name`
        may be used).

        accounts.acc_can_login
            **SELECT, FILTER, ORDER** | Logins enabled flag.

        accounts.acc_display_name
            **SELECT, FILTER, ORDER** | User display name.

        accounts.acc_email
            **SELECT, FILTER, ORDER** | Email address (RFC 2822).

        accounts.acc_full_name
            **SELECT, FILTER, ORDER** | User full name.

        accounts.acc_id
            **SELECT, FILTER, ORDER** | Local account identifier on system.

        accounts.acc_last_ip
            **SELECT, FILTER, ORDER** | Last IP address of user.

        accounts.acc_last_login
            **SELECT, FILTER, ORDER** | Timestamp of last account login.

        accounts.acc_last_ua
            **SELECT, FILTER, ORDER** | Last user agent string.

        accounts.acc_logins
            **SELECT, FILTER**        | Number of times user has logged in.

        accounts.acc_privs
            **SELECT, FILTER**        | Account privileges.

        accounts.acc_pw_change
            **SELECT, FILTER**        | Password change pending on next login.

        accounts.acc_session_time_started
            **SELECT, FILTER, ORDER** | Timestamp of session start time.

        accounts.acc_sort_name
            **SELECT, FILTER, ORDER** | User's sortable display name.

        accounts.acc_sys
            **SELECT, FILTER, ORDER** | System account flag.

        accounts.acc_time_created
            **SELECT, FILTER, ORDER** | Time account was created.

        accounts.acc_time_last_access
            **SELECT, FILTER, ORDER** | Time acocunts was last accessed.

        accounts.acc_type
            **SELECT, FILTER, ORDER** | Account type.

        accounts.acc_user_id
            **SELECT, FILTER, ORDER** | Local user ID on system.

        accounts.acc_uuid
            **SELECT, FILTER**        | Globally unique account identifier.

        accounts.com_id
            **SELECT, FILTER, ORDER** | Local community identifier on system.

        accounts.tz_id
            **SELECT, FILTER, ORDER** | Timezone identifier.

        communities.com_enabled
            **SELECT, FILTER, ORDER** | Community enabled flag.

        communities.com_full_name
            **SELECT, FILTER, ORDER** | Community name.

        communities.com_id
            **SELECT, FILTER, ORDER** | Local community identifier on system.

        communities.com_max_slots_per_res
            **SELECT, FILTER, ORDER** | Community maximum slots per reservation.

        communities.com_min_hours_btw_res
            **SELECT, FILTER, ORDER** | Community minimum hours between reservations.

        communities.com_mynetlab_news
            **SELECT, FILTER**        | Community MyNetlab page news markdown.

        communities.com_mynetlab_welcome
            **SELECT, FILTER**        | Community MyNetlab page welcome markdown.

        communities.com_tz_id
            **SELECT, FILTER, ORDER** | Timezone identifier.

        communities.com_uuid
            **SELECT, FILTER**        | Globally unique community identifier.

        timezones.tz_iana
            **SELECT, FILTER, ORDER** | Timezone IANA identifier.

        timezones.tz_id
            **SELECT, FILTER, ORDER** | Timezone identifier.

        timezones.tz_name
            **SELECT, FILTER, ORDER** | Timezone name.

        timezones.tz_sort
            **SELECT, FILTER, ORDER** | Timezone sort order.
        """

        total_pages = 1
        current_page = 1
        acc_ids = set()

        while current_page <= total_pages:
            data = await self.user_account_search(
                page=current_page,
                properties=properties,
                order=order,
                filter=filter,
                limit=limit,
            )
            current_page += 1
            total_pages = data['total_pages']
            for record in data['data']:
                acc_id = record['acc_id']
                if acc_id not in acc_ids:
                    acc_ids.add(acc_id)
                    yield record

    @minimum_version('17.1.4')
    async def user_account_search(
                self,
                *,
                properties: List[str] = ['acc_id'],
                order: Optional[Union[str, List[str]]] = None,
                filter=None,  # TODO
                limit: int = 100,
                page: Optional[int] = None,
                offset: Optional[int] = None,
                **kwargs
            ) -> Dict[str, Any]:
        """
        This method queries and/or retrieves user accounts. Also see
        :meth:`~netlab.api.UserApiMixin.user_account_search_iter`.

        Filters and returns paged results of user accounts. Any property from
        :meth:`~netlab.api.UserApiMixin.user_account_list` can be filtered upon. Multiple
        **filter** will be combined using AND. Returns zero or more user accounts that match filter query.

        :param properties: Properties consist of any combination of properties listed below under Properties.
        :param order: List of properties to sort by in order of preference.
        :type order: str or list
        :param filter: Search criteria. See below for how to use filters.
        :param limit: Page size or maximum number of records returned. This method can return a maximum of 500 use
            account records. Call the method multiple times with successive page numbers to return all data.
        :param page: Starting page number (pages start at one). When page is `None`, returns only info like
            'total_records' with no accounts. A page that goes pass all records will return no accounts.
        :param offset: Starting record number. If page and offset are both None, no records are returned; only meta
            data about total users and pages are returned.

        :return: A list of user accounts that match **filter**.

        **Examples**

        Collect the account id and account user id with **acc_id** that is 100031.

            Client().user_account_search(properties=['acc_id', 'acc_user_id'], page=1, filter={'acc_id': 100031})

        Collect first page of accounts, ordered by **acc_id**.

            Client().user_account_search(properties=['acc_id', 'acc_user_id'], page=1, order='acc_id')

        Return how many users can login.

            Client().user_account_search(filter={'acc_can_login': True})['total_records']

        **Properties**

        The following properties may be used in this method. The text in bold indicates the contexts in which the
        property may be used (see contexts below). The text following the | character describes the property.

        Properties in **communities** and **timezones** must be fully qualified (i.e. `timezones.tz_name`). Properties
        starting with **accounts** my be specified without namespace (i.e. `acc_full_name` or `accounts.acc_full_name`
        may be used).

        accounts.acc_can_login
            **SELECT, FILTER, ORDER** | Logins enabled flag.

        accounts.acc_display_name
            **SELECT, FILTER, ORDER** | User display name.

        accounts.acc_email
            **SELECT, FILTER, ORDER** | Email address (RFC 2822).

        accounts.acc_full_name
            **SELECT, FILTER, ORDER** | User full name.

        accounts.acc_id
            **SELECT, FILTER, ORDER** | Local account identifier on system.

        accounts.acc_last_ip
            **SELECT, FILTER, ORDER** | Last IP address of user.

        accounts.acc_last_login
            **SELECT, FILTER, ORDER** | Timestamp of last account login.

        accounts.acc_last_ua
            **SELECT, FILTER, ORDER** | Last user agent string.

        accounts.acc_logins
            **SELECT, FILTER**        | Number of times user has logged in.

        accounts.acc_privs
            **SELECT, FILTER**        | Account privileges.

        accounts.acc_pw_change
            **SELECT, FILTER**        | Password change pending on next login.

        accounts.acc_session_time_started
            **SELECT, FILTER, ORDER** | Timestamp of session start time.

        accounts.acc_sort_name
            **SELECT, FILTER, ORDER** | User's sortable display name.

        accounts.acc_sys
            **SELECT, FILTER, ORDER** | System account flag.

        accounts.acc_time_created
            **SELECT, FILTER, ORDER** | Time account was created.

        accounts.acc_time_last_access
            **SELECT, FILTER, ORDER** | Time acocunts was last accessed.

        accounts.acc_type
            **SELECT, FILTER, ORDER** | Account type.

        accounts.acc_user_id
            **SELECT, FILTER, ORDER** | Local user ID on system.

        accounts.acc_uuid
            **SELECT, FILTER**        | Globally unique account identifier.

        accounts.com_id
            **SELECT, FILTER, ORDER** | Local community identifier on system.

        accounts.tz_id
            **SELECT, FILTER, ORDER** | Timezone identifier.

        communities.com_enabled
            **SELECT, FILTER, ORDER** | Community enabled flag.

        communities.com_full_name
            **SELECT, FILTER, ORDER** | Community name.

        communities.com_id
            **SELECT, FILTER, ORDER** | Local community identifier on system.

        communities.com_max_slots_per_res
            **SELECT, FILTER, ORDER** | Community maximum slots per reservation.

        communities.com_min_hours_btw_res
            **SELECT, FILTER, ORDER** | Community minimum hours between reservations.

        communities.com_mynetlab_news
            **SELECT, FILTER**        | Community MyNetlab page news markdown.

        communities.com_mynetlab_welcome
            **SELECT, FILTER**        | Community MyNetlab page welcome markdown.

        communities.com_tz_id
            **SELECT, FILTER, ORDER** | Timezone identifier.

        communities.com_uuid
            **SELECT, FILTER**        | Globally unique community identifier.

        timezones.tz_iana
            **SELECT, FILTER, ORDER** | Timezone IANA identifier.

        timezones.tz_id
            **SELECT, FILTER, ORDER** | Timezone identifier.

        timezones.tz_name
            **SELECT, FILTER, ORDER** | Timezone name.

        timezones.tz_sort
            **SELECT, FILTER, ORDER** | Timezone sort order.
        """
        # Netlab does not understand bool; json.dumps(default) is not called for bools, only on
        # non-json-native objects.
        new_filter: Optional[Dict[str, Any]]
        if filter is None:
            new_filter = None
        else:
            new_filter = {}
            for k, v in filter.items():
                if isinstance(v, bool):
                    new_filter[k] = int(v)
                else:
                    new_filter[k] = v

        return await self.call(
            'user.account.search.task', properties=properties, order=order, filter=new_filter,
            limit=limit, page=page, offset=offset, **kwargs)

    async def user_account_get(
                self,
                *,
                acc_id: int,
                properties: Union[List[str], Literal['all', 'default']] = 'default',
                **kwargs
            ) -> Dict[str, Any]:
        """
        Retrieves a user account on the NETLAB system.

        :param acc_id: User Account ID.
        :param properties: Properties consist of any combination of properties listed below under Properties.
                           You may also use **"all"** to include all properties, or **"default"** for a smaller subset
                           of properties.

        :return: Account data for the requested acc_id with the requested properties.

        **Properties**

        acc_can_login
            User allowed to login.
        acc_display_name
            Account name people will see this name in discussions, messages and comments.
        acc_email
            Account email.
        acc_fs_id
            Filesystem identifier of the account root folder.
        acc_full_name
            Account full name.
        acc_id
            Account identifier.
        acc_last_ip
            Last IP address of user.
        acc_last_login
            Datetime of last account login.
        acc_last_ua
            User agent string of the browser when logging in from the UI.
        acc_logins
            Number of account logins.
        acc_privs
            Supplemental account privileges for instructors only. :py:class:`AccountPrivileges`
        acc_pw_change
            Force user to change password.
        acc_session_time_started
            Datetime of session start time.
        acc_sort_name
            Account name as it appears in sorted lists.
        acc_sys
            Mark as system account.
        acc_time_created
            Datetime account was created.
        acc_time_last_access
            Datetime of last access.
        acc_type
            Account type. :py:class:`AccountType`
        acc_user_id
            Unique system-wide login identifier.
        acc_uuid
            Globally unique account ID.
        com_full_name
            Community name.
        com_id
            Unique community identifier.
        tz_iana
            Timezone IANA database identifier.
        tz_id
            Time zone identifier.
        tz_name
            Time zone name.
        acc_nlx_terms
            Lab designer terms and conditions acknowleged.
        acc_npd_terms
            Pod designer terms and conditions acknowleged.
        acc_pw_console
            Instructor's lab device console password.
        acc_pw_enable
            Instructor's lab device enable password.
        accman_list_com_id
            Persist account manager community selection in list view (admin and syswide).
        date_format
            Date format preference. :py:class:`netlab.enums.DateFormat`
        time_format
            Time format preference. :py:class:`netlab.enums.TimeFormat`
        first_weekday
            First day of week preference. An integer from 0 - 6 mapping to Sunday - Saturday.
        page_length
            The number of items to display in paginated lists. Possible values are *10, 25, 50, 100*. A value of
            *-1* will disable pagination and return all items.
        """
        method = 'user.account.get'

        if properties == 'all':
            properties = self.user_account_get_props
        if properties == 'default':
            properties = self.user_account_default_props

        return await self.call(method, acc_id=acc_id, properties=properties, **kwargs)

    async def user_account_add(
                self,
                *,
                com_id: int,
                acc_user_id: str,
                acc_password: str,
                acc_full_name: str,
                acc_type: enums.AccountType = enums.AccountType.STUDENT,
                acc_privs: Optional[List[Union[str, enums.AccountPrivileges]]] = None,
                acc_display_name: Optional[str] = None,
                acc_sort_name: Optional[str] = None,
                acc_email: Optional[str] = None,
                acc_can_login: bool = True,
                acc_pw_change: bool = True,
                cls_id: Optional[int] = None,
                tz_id: int = 11,
                date_format: enums.DateFormat = enums.DateFormat.ISO,
                time_format: enums.TimeFormat = enums.TimeFormat.HOUR24,
                first_weekday: int = 0,
                page_length: int = 25,
                **kwargs
            ) -> int:
        """
        Add a user account on the NETLAB system. For accepted values see the properties documentation for
        :meth:`~netlab.api.UserApiMixin.user_account_get`.

        :param com_id: Community ID.
        :param acc_user_id: User Account ID.
        :param acc_password: Unencrypted account password.
        :param acc_full_name: Account full name.

        :param acc_type: Account type.
        :param acc_can_login: User can login. Defaults to True.
        :param acc_pw_change: Upon first login, force user to change password. Defaults to True.
        :param acc_privs: Supplemental account privileges for instructors in list or tuple
            format.
        :param acc_display_name: Account name, people will see this name in discussions, messages, and comments.
        :param acc_sort_name: Account name as it appears in sorted lists.
        :param acc_email: Account email address.
        :param cls_id: Class ID. (If you want to add user directly to a class during account creation.)
        :param tz_id: Timezone identifier. Defaults to 11.
        :param date_format: Date format preference.
        :param time_format: Time format preference.
        :param first_weekday: First day of week preference.
        :param page_length: The number of items to display in paginated lists.

        :return: On success, returns acc_id.
        """

        method = "user.account.add"
        params = {
            'com_id': com_id,
            'acc_user_id': acc_user_id,
            'acc_password': acc_password,
            'acc_full_name': acc_full_name,
            'acc_type': acc_type,
            'acc_can_login': acc_can_login,
            'acc_pw_change': acc_pw_change,
            'tz_id': tz_id,
            'date_format': date_format,
            'time_format': time_format,
            'first_weekday': first_weekday,
            'page_length': page_length
        }

        if acc_privs:
            acc_privs_str: List[str] = []
            for acc in acc_privs:
                if isinstance(acc, str):
                    acc_privs_str.append(acc)
                else:
                    acc_privs_str.append(acc.value)
            params['acc_privs'] = ','.join(acc_privs_str).upper()
        if acc_display_name:
            params['acc_display_name'] = acc_display_name
        if acc_sort_name:
            params['acc_sort_name'] = acc_sort_name
        if acc_email:
            params['acc_email'] = acc_email
        if cls_id:
            params['cls_id'] = cls_id

        params.update(kwargs)
        return int(await self.call(method, **params))

    async def user_account_password_set(
                self,
                *,
                acc_id: int,
                new_password: str,
                force_reset: bool = False,
                **kwargs
            ) -> Literal["OK"]:
        """
        Sets the user account password.

        :param acc_id: Account ID.
        :param new_password: New account password.
        :param force_reset: Force user to reset password.
        """

        method = "user.account.password.set"
        return await self.call(method, acc_id=acc_id, new_password=new_password, force_reset=force_reset, **kwargs)

    async def user_account_remove(
                self,
                *,
                acc_id: int,
                **kwargs
            ) -> None:
        """
        Removes the specified user account.

        :param acc_id: User Account ID
        """

        method = "user.account.remove"
        return await self.call(method, acc_id=acc_id, **kwargs)

    async def user_account_update(
                self,
                *,
                acc_id,
                acc_user_id=None,
                acc_type=None,
                acc_can_login=None,
                acc_pw_change=None,
                acc_password=None,
                acc_privs=None,
                acc_full_name=None,
                acc_display_name=None,
                acc_sort_name=None,
                acc_email=None,
                tz_id=None,
                date_format=None,
                time_format=None,
                first_weekday=None,
                page_length=None,
                **kwargs
            ) -> Literal["OK"]:
        """
        Update the user account. For accepted values see the properties documentation for
        :meth:`~netlab.api.UserApiMixin.user_account_get`.

        :param int acc_id: Account ID

        :param str acc_user_id: Unique system-wide login identifier.
        :param enums.AccountType acc_type: Account type.
        :param bool acc_can_login: User allowed to login.
        :param bool acc_pw_change: Force user to change password.
        :param str acc_password: Account password
        :param list[enums.AccountPrivileges] acc_privs: Supplemental account privileges for instructors in
            list or tuple format.
        :param str acc_full_name: Account full name.
        :param str acc_display_name: Account name people will see this name in discussions, messages and comments.
        :param str acc_sort_name: Account name as it appears in sorted lists.
        :param str acc_email: Account e-mail address.
        :param int tz_id: Timezone identifier.
        :param DateFormat date_format: Date format preference.
        :param TimeFormat time_format: Time format preference.
        :param int first_weekday: First day of week preference.
        :param int page_length: The number of items to display in paginated lists.
        """
        method = "user.account.update"
        params: Dict[str, Any] = {'acc_id': acc_id}

        if acc_user_id:
            params['acc_user_id'] = acc_user_id
        if acc_type:
            params['acc_type'] = acc_type
        if acc_can_login is not None:
            params['acc_can_login'] = acc_can_login
        if acc_pw_change is not None:
            params['acc_pw_change'] = acc_pw_change
        if acc_password:
            params['acc_password'] = acc_password
        if acc_privs:
            if isinstance(acc_privs, (list, tuple, set)):
                acc_privs = ",".join(map(lambda v: v.value, acc_privs))
            params['acc_privs'] = acc_privs.upper()
        if acc_full_name:
            params['acc_full_name'] = acc_full_name
        if acc_display_name:
            params['acc_display_name'] = acc_display_name
        if acc_sort_name:
            params['acc_sort_name'] = acc_sort_name
        if acc_email:
            params['acc_email'] = acc_email
        if tz_id:
            params['tz_id'] = tz_id
        if date_format:
            params['date_format'] = date_format
        if time_format:
            params['time_format'] = time_format
        if first_weekday:
            params['first_weekday'] = first_weekday
        if page_length:
            params['page_length'] = page_length

        params.update(kwargs)
        return await self.call(method, **params)

    async def user_community_add(
                self,
                *,
                com_full_name: str,
                com_id: int,
                com_enabled: bool = True,
                com_max_slots_per_res: Optional[int] = None,
                com_min_hours_btw_res: Optional[int] = None,
                com_mynetlab_news: Optional[str] = None,
                com_mynetlab_welcome: Optional[str] = None,
                **kwargs
            ) -> int:
        """
        This method allows you to add a community to the NETLAB+ system.

        :param com_full_name: Community full name.
        :param com_id: Community identifier.
        :param com_enabled: Is community enabled.
        :param com_max_slots_per_res: Max amount of slots per reservation.
        :param com_min_hours_btw_res: Minimum hours between reservations.
        :param com_mynetlab_news: Community MyNETLAB+ news entry.
        :param com_mynetlab_welcome: Community MyNETLAB+ welcome message.

        :return: Community ID of the newly created community.
        """

        assert com_id < 100, "Community id out of range"  # TODO document and confirm
        return await self.call(
            'user.community.add',
            com_full_name=com_full_name,
            com_id=com_id,
            com_enabled=com_enabled,
            com_max_slots_per_res=com_max_slots_per_res,
            com_min_hours_btw_res=com_min_hours_btw_res,
            com_mynetlab_news=com_mynetlab_news,
            com_mynetlab_welcome=com_mynetlab_welcome,
            **kwargs)

    async def user_community_update(
                self,
                *,
                com_id: int,
                com_full_name: Optional[str] = None,
                com_enabled: Optional[bool] = None,
                com_mynetlab_welcome: Optional[str] = None,
                com_mynetlab_news: Optional[str] = None,
                com_min_hours_btw_res: Optional[int] = None,
                com_max_slots_per_res: Optional[int] = None,
                com_ext_slots_per_res: Union[None, int, enums.CommunityExtensionSlots] = None,
                **kwargs
            ) -> None:
        """
        Update a community.

        :param int com_id: Community ID.
        :param str com_full_name: Community full name.
        :param bool com_enabled: Is community enabled.
        :param str com_mynetlab_welcome: Community MyNETLAB+ welcome message.
        :param str com_mynetlab_news: Community MyNETLAB+ news entry.
        :param int com_min_hours_btw_res: Minimum hours between reservations.
        :param int com_max_slots_per_res: Max amount of slots per reservation.
        :param com_ext_slots_per_res: Allow extensions to reservations for students and teams.
            Instructors are not restricted to this limit.
        :type com_ext_slots_per_res: int or enums.CommunityExtensionSlots
        """
        return await self.call(
            'user.community.update', com_id=com_id, com_full_name=com_full_name, com_enabled=com_enabled,
            com_mynetlab_welcome=com_mynetlab_welcome, com_mynetlab_news=com_mynetlab_news,
            com_min_hours_btw_res=com_min_hours_btw_res, com_max_slots_per_res=com_max_slots_per_res,
            com_ext_slots_per_res=com_ext_slots_per_res, **kwargs)

    async def user_community_remove(self, *, com_id, **kwargs) -> None:
        """
        This method allows you to remove a community.

        :param int com_id: Community identifier.
        """

        return await self.call('user.community.remove', com_id=com_id)

    async def user_community_find(self, *, com_full_name: str, **kwargs) -> Optional[int]:
        """
        This method allows you to find a community based on **com_full_name**.

        :param com_full_name: Community full name.

        :return: The Community ID on success.
        """

        method = "user.community.find"
        result = await self.call(method, com_full_name=com_full_name, **kwargs)

        return None if result is None else int(result)

    async def user_community_get(
                self,
                *,
                com_id: int,
                properties: Union[List[str], Literal['all', 'default']] = "default",
                **kwargs
            ) -> Dict[str, Any]:
        """
        This method allows you to retrieve a community and its properties.

        :param int com_id: Community ID.
        :param properties: Properties to return. Accepts string or list.
                        See below for all available properties. You may also use **"all"** to include all properties.
        :type properties: list[str] or 'all'

        :return: An object with the properties requested:
        :rtype: dict

        **Properties**

        com_alt_banner
            Alternate community banner.
        com_alt_banner_height
            Alternate banner height.
        com_alt_banner_width
            Alternate banner width.
        com_enabled
            Is community enabled.
        com_full_name
            Community full name.
        com_id
            Community identifier.
        com_max_slots_per_res
            Max amount of slots per reservation.
        com_min_hours_btw_res
            Minimum hours between reservations.
        com_mynetlab_news
            Community MyNETLAB+ news entry.
        com_mynetlab_welcome
            Community MyNETLAB+ welcome message.
        """

        method = "user.community.get"
        if properties == 'all':
            properties = self.community_props
        if properties == 'default':
            properties = self.community_default_props

        return await self.call(method, com_id=com_id, properties=properties, **kwargs)

    async def user_community_list(self, *, properties="default", sort_property="com_id", **kwargs):
        """
        Retrieve a list of communities.

        :param properties: Properties to return. Accepts string or list.
                        See below for all available properties. You may also use **"all"** to include all properties.
        :type properties: list[str] or 'all' or 'default'
        :param str sort_property: Property to sort list by. Only **com_id** (default) and **com_full_name** are
                   valid options.

        :return: A list of objects with the properties requested:
        :rtype: list[dict]

        **Properties**

        com_alt_banner
            Alternate community banner.
        com_alt_banner_height
            Alternate banner height.
        com_alt_banner_width
            Alternate banner width.
        com_enabled
            Is community enabled.
        com_full_name
            Community full name.
        com_id
            Community identifier.
        com_max_slots_per_res
            Max amount of slots per reservation.
        com_min_hours_btw_res
            Minimum hours between reservations.
        com_mynetlab_news
            Community MyNETLAB+ news entry.
        com_mynetlab_welcome
            Community MyNETLAB+ welcome message.
        """

        method = "user.community.list"
        if properties == 'all':
            properties = self.community_props
        if properties == 'default':
            properties = self.community_default_props

        return await self.call(method, properties=properties, sort_property=sort_property, **kwargs)

    async def user_logins_system_get(self, **kwargs):
        """
        This method allows you to query the User Logins status and number of active logins.

        :return: A dictionary with the following properties:
        :rtype: dict

        active
            Number of active users logged in.
        enabled
            User logins status.
        """
        method = "user.logins.system.get"
        return await self.call(method, **kwargs)

    async def user_logins_system_set(self, *, enabled, logout=False, **kwargs):
        """
        This method allows you to enable and disable user logins.

        :param bool enabled: Enable or disable user logins.
        :param bool logout: Logout users that are already logged in.

        :return: A dictionary with the following properties:
        :rtype: dict

        changed
            Whether user login status was changed.
        logouts
            Number of users logged out.
        """
        method = "user.logins.system.set"
        return await self.call(method, enabled=enabled, logout=logout, **kwargs)

    @minimum_version('18.2.0')
    async def user_session_keepalive(self, *, acc_id, **kwargs):
        """
        This method performs a keepalive on a logged in user's session and
        restarts their session timeout. The last access for the session is set
        to the current time.

        :param int acc_id: Local account identifier.

        :return: A dictionary with the following properties:
        :rtype: dict

        acc_session_time_started
            Session start time, or None if the administrator specifies an
            acc_id that is not logged in.
        acc_time_last_access
            Updated last access time, or None if the administrator specifies an
            acc_id that is not logged in.
        """
        method = "user.session.keepalive"
        return await self.call(method, acc_id=acc_id, **kwargs)



import asyncio
from datetime import datetime, timedelta
from typing import Union


class Throttler(object):
    """
    BETA: Provides throttling using the Leaky Bucket Algorithm.

    https://en.wikipedia.org/wiki/Leaky_bucket

    It can be imagined like so:

    There are a number of buckets, full of water. There is a tap, that can fill
    a bucket in (refresh) seconds. For a task to start, it must flip a filled
    bucket. It will keep the bucket overturned until it completes. When it
    completes, the bucket will wait to be filled so that another task can use
    it.

    Note that `refresh = 0` is just a semaphore.

    :param pod_id: Pod ID
    :param state: PC Pod State
    """

    _restore_time: timedelta
    _restore_last: datetime
    _buckets_total: float
    _buckets_taken: float
    _buckets_pending: float
    _buckets_are_pending: asyncio.Event
    _lock: asyncio.Lock

    def __init__(
        self,
        buckets: float,
        refresh: Union[float, timedelta],
    ):
        """
        Create a throttle object.

        :param buckets: Number of buckets (concurrent tasks).
        :param refresh: Time (seconds) taken to fill a bucket.
        """

        if isinstance(refresh, timedelta):
            self._restore_time = refresh
        else:
            self._restore_time = timedelta(seconds=refresh)
        self._restore_last = datetime.utcnow()
        self._buckets_total = buckets
        self._buckets_taken = 0.0
        self._buckets_pending = 0.0
        self._buckets_are_pending = asyncio.Event()
        self._buckets_are_pending.set()
        self._lock = asyncio.Lock()

    def _refresh(self):
        now = datetime.utcnow()
        restored_buckets = (now - self._restore_last) / self._restore_time
        self._buckets_pending -= restored_buckets
        self._buckets_pending = max(0, self._buckets_pending)
        self._restore_last = now
        if self._buckets_pending + self._buckets_taken < self._buckets_total:
            return 0
        else:
            taken = self._buckets_pending
            time = taken % 1 * self._restore_time
            return time.total_seconds()

    async def _grab_bucket(self):
        async with self._lock:
            if self._buckets_pending == 0:
                await self._buckets_are_pending.wait()
                self._buckets_are_pending.clear()
            wait_time = self._refresh()
            await asyncio.sleep(wait_time)
            self._buckets_taken += 1

    def _release_bucket(self):
        self._refresh()
        self._buckets_taken -= 1
        self._buckets_pending += 1
        self._buckets_are_pending.set()

    async def wrap(self, co):
        try:
            await self._grab_bucket()
        except asyncio.CancelledError as e:
            await co.throw(e)
        try:
            result = await co
            return result
        finally:
            self._release_bucket()

    async def __aenter__(self) -> None:
        await self._grab_bucket()

    async def __aexit__(self, exc_type, exc_value, traceback):
        self._release_bucket()



from typing import Any, List, Literal, Optional, TypedDict, Union

from typing_extensions import Required

from ._client_protocol import ClientProtocol


class LabExerciseListResult(TypedDict, total=False):
    ex_id: Required[str]
    "Lab exercise identifier."
    ex_con_id: Required[str]
    "Lab design identifier."
    ex_im_id: Required[Union[str, None]]
    "Internal image map identifier for clickable hotspots."
    ex_index: str
    "Exercise index number in lab design."
    ex_name: str
    "Exercise name."
    ex_content_url: str
    "Lab document relative URL."
    ex_minutes: int
    "Exercise minimum time in minutes."
    ex_topology_image: Union[str, None]
    "Lab topology image relative URL."
    ex_vlan_map: Union[str, None]
    "Alternate VLAN map for real equipment pods, overrides pod design."
    ex_pt_id: str
    "Exercise pod type identifier."
    ex_tabs: str
    "Exercise tabs enabled."
    ex_actions: str
    "Exercise actions enabled."
    ex_fs_id: str
    "Root node of folder containing configuration files for this exercise."
    ex_content_no_preview: bool
    "If true, user cannot preview this exercise."
    ex_ilt_no_select: int
    "If true, exercise is disabled in ILT."
    cls_ids: List[int]
    "Array of class identifiers using this content."
    hotspots: List[Any]
    "Array of image_map objects (for clickable hotspots)."


class LabApiMixin(ClientProtocol):
    "Methods for working with labs."

    async def lab_exercise_list(
        self,
        *,
        properties: Optional[List[str]] = None,
        sort_property: Literal["ex_id", "ex_index", "ex_name", "ex_con_id"] = "ex_id",
        con_id: Optional[str] = None,
        **kwargs
    ) -> List[LabExerciseListResult]:
        """
        This method allows you to retrieve a list of lab exercises installed on the system.

        :param con_id: If specified, only return exercises within for the lab designed with identifier of con_id.
        :param sort_property: Property to sort results by.
        :param properties: List of properties to retrieve. See Properties section for list of available properties.

        :return: On success, returns an object with the requested properties:

        **Properties**

        ex_id
            Lab exercise identifier.
        ex_con_id
            Lab design identifier.
        ex_index
            Exercise index number in lab design.
        ex_name
            Exercise name.
        ex_content_url
            Lab document relative URL.
        ex_minutes
            Exercise minimum time in minutes.
        ex_topology_image
            Lab topology image relative URL.
        ex_im_id
            Internal image map identifier for clickable hotspots.
        ex_vlan_map
            Alternate VLAN map for real equipment pods, overrides pod design.
        ex_pt_id
            Exercise pod type identifier.
        ex_tabs
            Exercise tabs enabled.
        ex_actions
            Exercise actions enabled.
        ex_fs_id
            Root node of folder containing configuration files for this exercise.
        ex_content_no_preview
            If true, user cannot preview this exercise.
        ex_ilt_no_select
            If true, exercise is disabled in ILT.
        cls_ids
            Array of class identifiers using this content.
        hotspots
            Array of image_map objects (for clickable hotspots).
        """
        method = "lab.exercise.list"

        return await self.call(
            method,
            con_id=con_id,
            sort_property=sort_property,
            properties=properties,
            **kwargs
        )



"""
Config can be passed to :py:class:`netlab.async_client.NetlabClient` using environment variables,
using a json file on your system, or by passing a ``dict`` explicitly.
"""

import ssl
from enum import Enum
from typing import Literal, Optional, TypedDict, Union, cast

RAW_CIPHER_LIST = Union[None, "NetlabCipherListEnum", str]


class NetlabCipherListEnum(Enum):
    "Preconfigured Netlab cipher lists."

    NETLAB = "NETLAB"
    PYTHON = "PYTHON"
    "Default netlab cipher list."

    @classmethod
    def get_cipher_list(cls, ciphers: RAW_CIPHER_LIST) -> Optional[str]:
        try:
            ciphers = cls[cast(str, ciphers)]
        except KeyError:
            pass

        if ciphers is cls.NETLAB:
            return "DEFAULT:AES256-GCM-SHA384"
        if ciphers is cls.PYTHON:
            return None
        else:
            return ciphers


class NetlabServerConfig(TypedDict):
    """
    Passed to `netlab.async_client.NetlabClient` to configure the connection
    to a server.

    **Example**

    ::

        config = {
            "host": "192.168.1.10",
            "user": "administrator",
            "token": "S25GWP5P2247CMRDNLTCNKATT49KSGEDDPXMTM6A",
            "port": 9000,
        }

        async with NetlabClient(config=config) as client:
            ...
    """

    host: str
    "IP Address or hostname of remote NETLAB+ API."
    user: str
    "Currently ``'administrator'`` is the only user account allowed to use the API."
    token: str
    "Token obtained from NETLAB+ system."
    port: int
    "Port of remote NETLAB+ API. Usually ``9000``."
    server_hostname: Optional[str]
    "The hostname that ssl is configured to use. Only required if it is diffrent from 'host'."
    ssl: Union[Literal["default", "noverify"], ssl.SSLContext]
    "You may pass `ssl.SSLContext` for complete control over ssl. 'self_signed' and 'noverify' are equivilant."
    ssl_ciphers: RAW_CIPHER_LIST
    "ssl ciphers to use. This can be a value from `NetlabCipherListEnum` or an openssl cipher list."
    message_byte_limit: int
    "The maximum size of a NETLAB+ message."


class NetlabServerEnvConfig(TypedDict, total=True):
    """
    Each of the above configuration options can be specified as environment variables.
    """

    NETLAB_CONFIG: str
    "Folder which contains the 'config.json'."
    NETLAB_CONFIG_HOST: str
    "IP Address or hostname of remote NETLAB+ API."
    NETLAB_CONFIG_USER: str
    "User account ID from NETLAB+ system."
    NETLAB_CONFIG_TOKEN: str
    "Token obtained from NETLAB+ system."
    NETLAB_CONFIG_PORT: str
    "Port of remote NETLAB+ API."
